// io.cpp
//

//#undef BOCHS  // In ambiente Windows decommentare la definizione
//#define BOCHS  //  giusta a seconda che si voglia generare uma 
   //  immagine funzionante (porta seriale e 
   //  riconoscimento hard disk) con l'emulatore Bochs 
   //  o con un PC reale.
   //  Se si dispone di make e` sufficiente $make BOCHS=no
   //  (vedere il README)

//funzioni di libreria
static void itostr(char *buf, unsigned int len, long l);

////////////////////////////////////////////////////////////////////////////////
//                                 COSTANTI                                   //
////////////////////////////////////////////////////////////////////////////////

// Livelli di privilegio dei processi
// Devono coincidere con i valori in sistema.cpp
//
const char LIV_UTENTE = 3, LIV_SISTEMA = 0;

// Interruzione hardware a priorita' minore 
//
const int IRQ_MAX = 15;

// Priorita' base dei processi esterni
// La priorita' di un processo esterno e' calcolata aggiungendo a questa
//  costante la priorita' dell' interruzione a cui e' associato (la priorita'
//  delle interruzioni e' inversamente proporzionale al loro tipo, quindi si
//  somma IRQ_MAX - TIPO alla priorita' di base).
// Deve coincidere con il valore in sistema.cpp
//
const int PRIO_ESTERN_BASE = 0x400;

// Identificatori dei processi esterni
// Devono coincidere con i valori in sistema.cpp
//
enum estern_id {
 tastiera,
 com1_in,
 com1_out,
 com2_in,
 com2_out,
 floppy,
 ata0,
 ata1
};

typedef char* ind_b;  // indirizzo di un byte (una porta)
typedef short* ind_w;  // indirizzo di una parola
typedef int* ind_l;  // indirizzo di una parola lunga
typedef unsigned int ind_fisico;// locazione di memoria fisica (per il DMA)
typedef unsigned int size_t;

////////////////////////////////////////////////////////////////////////////////
//                        CHIAMATE DI SISTEMA USATE                           //
////////////////////////////////////////////////////////////////////////////////

extern "C" void terminate_p(void);
extern "C" void sem_ini(int &index_des_s, int val, bool &risu);
extern "C" void sem_wait(int sem);
extern "C" void sem_signal(int sem);

/*
void delay(int ms)
{
 for(int i=0;i<(ms*3+1);i++)
 {}
}
*/

// Attende N interruzioni del timer di sistema
//
extern "C" void delay(int n);

////////////////////////////////////////////////////////////////////////////////
//               INTERFACCIA OFFERTA DAL NUCLEO AL MODULO DI IO               //
////////////////////////////////////////////////////////////////////////////////

// Alloca DIM bytes nella memoria dinamica
//
extern "C" void *mem_alloc(int dim);

// Rilascia una zona di memoria indirizzata da PV precedentemente allocata
//  con mem_alloc()
//
extern "C" void mem_free(void *pv);

extern "C" void activate_pe(void f(int), int a, int prio, char liv,
 short &identifier, estern_id interf, bool &risu);

enum controllore { master=0, slave=1 };
extern "C" void nwfi(controllore c);

extern "C" bool verifica_area(void *area, unsigned int dim, bool write);
extern "C" void trasforma(ind_l vetti, ind_fisico &iff);
extern "C" void panic(const char *msg);
//extern "C" void dbg(const char *msg);
extern "C" void dbg_int(const unsigned int num);
extern "C" void fill_gate(int gate, void (*f)(void), int tipo, int dpl);
extern "C" void reboot(void);

//
// Chiamate per l' IO da console
//

const int CON_BUF_SIZE = 0x0fa0;

struct con_status {
 unsigned char buffer[CON_BUF_SIZE];
 short x, y;
};

extern "C" void con_read(char &ch, bool &risu);
extern "C" void con_write(const char *vett, int quanti);

extern "C" void con_save(con_status *cs);
extern "C" void con_load(const con_status *cs);
extern "C" void con_update(con_status *cs, const char *vett, int quanti);
extern "C" void con_init(con_status *cs);

////////////////////////////////////////////////////////////////////////////////
//                      FUNZIONI GENERICHE DI SUPPORTO                        //
////////////////////////////////////////////////////////////////////////////////


// protezione dalla concorrenza con le interruzioni
//
inline void lock() { asm("cli"); }
inline void unlock() { asm("sti"); }

// ingresso di un byte da una porta di IO
extern "C" void inputb(ind_b reg, char &a);

// uscita di un byte su una porta di IO
extern "C" void outputb(char a, ind_b reg);

// ingresso di una word da una porta di IO
extern "C" void inputw(ind_b reg, short &a);

// uscita di una word su una porta di IO
extern "C" void outputw(short a, ind_b reg);

// ingresso di una word da una porta di IO
extern "C" void inputl(ind_b reg, int &a);

// uscita di una word su una porta di IO
extern "C" void outputl(int a, ind_b reg);

// ingresso di una stringa di word da un buffer di IO
extern "C" void inputbuffw(ind_w reg, short *a,short n);

// uscita di una stringa di word su un buffer di IO
extern "C" void outputbuffw(short *a, ind_w reg,short n);

void *memcpy(void *dest, const void *src, unsigned int n);
void *memset(void *dest, int c, size_t n);

////////////////////////////////////////////////////////////////////////////////
//                    GESTIONE DELLE INTERFACCE SERIALI                       //
////////////////////////////////////////////////////////////////////////////////

const int S = 2;

extern "C" void go_inputse(ind_b i_ctr);
extern "C" void halt_inputse(ind_b i_ctr);
extern "C" void go_outputse(ind_b i_ctr);
extern "C" void halt_outputse(ind_b i_ctr);

enum funz { input_n, input_ln, output_n, output_0 };

struct interfse_reg {
 ind_b iRBR;
 ind_b iTHR;
 ind_b iLSR;
 ind_b iIER;
 ind_b iIIR;
};

struct des_se {
 interfse_reg indreg;
	int mutex;
	int sincr;
	int cont;
	ind_b punt;
	funz funzione;
	char stato;
};

extern "C" des_se com[S];

void startse_in(des_se *p_des, char vetti[], int quanti, funz op)
{
	p_des->cont = quanti;
	p_des->punt = vetti;
	p_des->funzione = op;
 go_inputse(p_des->indreg.iIER);
}

extern "C" void c_readse_n(int serial, char vetti[], int quanti, char &errore)
{
 des_se *p_des;

 if(serial < 0 || serial >= S || !verifica_area(vetti, quanti, true) ||
   !verifica_area(&errore, sizeof(char), true))
  return;

 p_des = &com[serial];
 sem_wait(p_des->mutex);
 startse_in(p_des, vetti, quanti, input_n);
 sem_wait(p_des->sincr);
 errore = p_des->stato;
 sem_signal(p_des->mutex);
}

extern "C" void c_readse_ln(int serial, char vetti[], int &quanti, char &errore)
{
 des_se *p_des;

 if(serial < 0 || serial >= S ||
   !verifica_area(&quanti, sizeof(int), true) ||
   !verifica_area(vetti, quanti, true) ||
    !verifica_area(&errore, sizeof(char), true))
  return;

 p_des = &com[serial];
 sem_wait(p_des->mutex);
 startse_in(p_des, vetti, 80, input_ln);
 sem_wait(p_des->sincr);
 quanti = p_des->cont;
 errore = p_des->stato;
 sem_signal(p_des->mutex);
}

void input_com(int h)
{
 char c, s;
 bool fine;
 des_se *p_des;

 p_des = &com[h];

 for(;;) {
  fine = false;
  halt_inputse(p_des->indreg.iIER);
  inputb(p_des->indreg.iRBR, c);
  inputb(p_des->indreg.iLSR, s);

  p_des->stato = s&0x1e;
  if(p_des->stato != 0)
   fine = true;
  else {
   if(p_des->funzione == input_n) {
    *p_des->punt = c;
    p_des->punt++;
    p_des->cont--;
    if(p_des->cont == 0)
     fine = true;
   } else if(p_des->funzione == input_ln)
    if(c == '\r' || c == '\n') {
     fine = true;
     p_des->cont = 80 - p_des->cont;
    } else {
     *p_des->punt = c;
     p_des->punt++;
     p_des->cont--;
     if(p_des->cont == 0) {
      fine = true;
      p_des->cont = 80;
     }
    }
  }

		if(fine == true) {
			*p_des->punt = 0;	// manca *
			sem_signal(p_des->sincr);
		} else
			go_inputse(p_des->indreg.iIER);

		nwfi(master); // sia com1 che com2 sono sul master
	}
}

// forza l' output del primo byte, sembra che l' interfaccia emulata da
//  Bochs presenti la richiesta di interruzione solo quando THR si svuota
//  dopo l' invio del primo byte
//
void kickse_out(des_se *p_des)
{
	char c = *p_des->punt;
	bool fine = false;

	p_des->punt++;

	if(p_des->funzione == output_n) {
		p_des->cont--;
		if(p_des->cont == 0) {
			halt_outputse(p_des->indreg.iIER);
			fine = true;
		}

		outputb(c, p_des->indreg.iTHR);
	} else if(p_des->funzione == output_0) {
		if(c == 0) {
			fine = true;
			halt_outputse(p_des->indreg.iIER);
		} else {
			p_des->cont++;
			outputb(c, p_des->indreg.iTHR);
		}
	}

	if(fine == true)
		sem_signal(p_des->sincr);
}

void startse_out(des_se *p_des, char vetto[], int quanti, funz op)
{
	p_des->cont = quanti;
	p_des->punt = vetto;
	p_des->funzione = op;
	go_outputse(p_des->indreg.iIER);
#ifdef BOCHS
 kickse_out(p_des);
#endif
}

extern "C" void c_writese_n(int serial, char vetto[], int quanti)
{
 des_se *p_des;

 if(serial < 0 || serial >= S || !verifica_area(vetto, quanti, false))
  return;

 p_des = &com[serial];
 sem_wait(p_des->mutex);
 startse_out(p_des, vetto, quanti, output_n);
 sem_wait(p_des->sincr);
 sem_signal(p_des->mutex);
}

extern "C" void c_writese_0(int serial, char vetto[], int &quanti)
{
 des_se *p_des;

 if(serial < 0 || serial >= S ||
   !verifica_area(&quanti, sizeof(int), true) ||
   !verifica_area(vetto, quanti, false))
  return;

 p_des = &com[serial];
 sem_wait(p_des->mutex);
 startse_out(p_des, vetto, 0, output_0);
 sem_wait(p_des->sincr);
 quanti = p_des->cont;
 sem_signal(p_des->mutex);
}

void output_com(int h)
{
 char c;
 bool fine;
 des_se *p_des;

 p_des = &com[h]; // como

	for(;;) {
		fine = false;
		if(p_des->funzione == output_n) {
			p_des->cont--;
			if(p_des->cont == 0) {
				fine = true;
				halt_outputse(p_des->indreg.iIER);
			}

			c = *p_des->punt;
			outputb(c, p_des->indreg.iTHR);
			p_des->punt++;
		} else if(p_des->funzione == output_0) {
			c = *p_des->punt;
			if(c == 0) {
				fine = true;
				halt_outputse(p_des->indreg.iIER);
			} else {
				outputb(c, p_des->indreg.iTHR);
				p_des->cont++;
				p_des->punt++;
			}
		}

		if(fine == true)
			sem_signal(p_des->sincr);

		nwfi(master); // sia com1 che com2 sono sul master
	}
}

extern "C" void com_setup(void);

const estern_id com_id[2][2] = {
	{ com1_in, com1_out },
	{ com2_in, com2_out }
};

// Interruzione hardware della seconda interfaccia seriale
//
const int COM2_IRQ = 4;

void com_init()
{
	des_se *p_des;
	bool r1, r2;
	short id;
	int i, com_base_prio = PRIO_ESTERN_BASE + IRQ_MAX - COM2_IRQ;

	com_setup();

	for(i = 0; i < S; ++i) {
		p_des = &com[i];

		sem_ini(p_des->mutex, 1, r1);
		sem_ini(p_des->sincr, 0, r2);
		if(!r1 || !r2)
			panic("Impossibile allocare semaforo per l' IO");

		activate_pe(input_com, i, com_base_prio - i,
			LIV_SISTEMA, id, com_id[i][0], r1);
		activate_pe(output_com, i, com_base_prio - i,
			LIV_SISTEMA, id, com_id[i][1], r2);
		if(!r1 || !r2)
			panic("Impossibile creare processo esterno di uscita");
	}
}

////////////////////////////////////////////////////////////////////////////////
//                         GESTIONE DELLA TASTIERA                            //
////////////////////////////////////////////////////////////////////////////////

#define KBD_MOD_NR	2		// numero di modificatori usati
#define KBD_COD_NR	127		// numero di codici

// modificatori
enum kbd_mod {
	KBD_MOD_NONE = 0,
	KBD_MOD_SHIFT = 1,
	KBD_MOD_ALT = 2,
	KBD_MOD_ALTGR = 3,
	KBD_MOD_ALTSH = 4,
	KBD_MOD_CTRL = 5
};

#define KBD_SHIFT	(KBD_MOD_SHIFT << 8)
#define KBD_ALT		(KBD_MOD_ALT << 8)
#define KBD_ALTGR	(KBD_MOD_ALTGR << 8)
#define KBD_ALTSH	(KBD_MOD_ALTSH << 8)
#define KBD_CTRL	(KBD_MOD_CTRL << 8)

#define KBD_MOD_MASK	0x0f00

#define KBD_IS_MOD(c)	((c&KBD_MOD_MASK) != 0)
#define KBD_MOD(c)	((c&KBD_MOD_MASK) >> 8)

// tasti funzione
#define KBD_F1		0x1000
#define KBD_F2		0x2000
#define KBD_F3		0x3000
#define KBD_F4		0x4000
#define KBD_F5		0x5000
#define KBD_F6		0x6000
#define KBD_F7		0x7000
#define KBD_F8		0x8000
#define KBD_F9		0x9000
#define KBD_F10		0xa000
#define KBD_F11		0xb000
#define KBD_F12		0xc000

#define KBD_FNUM(code)	(((code) - KBD_F1) >> 12)

#define KBD_CANC	0xd000

// tabella di corrispondenza tra codici e caratteri ascii (tastiera italiana)
//
static unsigned short kbd_keymap[KBD_MOD_NR][KBD_COD_NR] = {
	{ 0, 27, '1', '2', '3', '4', '5', '6', '7', '8',
	'9', '0', '\'', 141, '\b', '\t', 'q', 'w', 'e', 'r',
	't', 'y', 'u', 'i', 'o', 'p', 138, '+', '\n', KBD_CTRL,
	'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', 149,
	133, '\\', KBD_SHIFT, 151, 'z', 'x', 'c', 'v', 'b', 'n',
	'm', ',', '.', '-', KBD_SHIFT, '*', KBD_ALT, ' ', 0, KBD_F1,
	KBD_F2, KBD_F3, KBD_F4, KBD_F5, KBD_F6, KBD_F7, KBD_F8, KBD_F9,
		KBD_F10, 0,
	0, 0, 0, 0, '-', 0, 0, 0, '+', 0,
	0, 0, 0, KBD_CANC, '\n', 0, 0, KBD_F11, KBD_F12, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0 },

	{ 0, 27, '!', '"', 156, '$', '%', '&', '/', '(',
	')', '=', '?', '^', 0, 0, 'Q', 'W', 'E', 'R',
	'T', 'Y', 'U', 'I', 'O', 'P', 130, '*', 0, KBD_CTRL,
	'A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', 135,
	248, '|', KBD_SHIFT, 21, 'Z', 'X', 'C', 'V', 'B', 'N',
	'M', ';', ':', '_', KBD_SHIFT, '*', KBD_ALT, ' ', 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, '7', '8', '9', '-', '4', '5', '6', '+', '1',
	'2', '3', '0', '.', 0, 0, '>', 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0 }
};

// dimensione della coda di una tastiera
const int KBD_QUEUE_SIZE = 512;

// maschere per il campo flags
enum {
	KBD_FLAG_NONE = 0,
	KBD_FLAG_SHIFT = 0x100,
	KBD_FLAG_ALT = 0x200,
	KBD_FLAG_CTRL = 0x400
};

struct des_term;

// tastiera virtuale
struct kbd_t {
	char queue[KBD_QUEUE_SIZE];
	int in, out;
	des_term *term;
	int flags;
	int nl_cnt;
};

// flag in corrispondenza dei modificatori
const int kbd_modflag_tab[6] = {
	KBD_FLAG_NONE,
	KBD_FLAG_SHIFT,
	KBD_FLAG_ALT,
	KBD_FLAG_ALT,
	KBD_FLAG_NONE,
	KBD_FLAG_CTRL
};

// tastiera virtuale attiva
static kbd_t *kbd_att = 0;

// inizializzazione di una tastiera virtuale
inline void kbd_init(kbd_t *kbd, des_term *term)
{
	memset(kbd, 0, sizeof(kbd_t));
	kbd->term = term;
	if(!kbd_att)
		kbd_att = kbd;
}

// vera se e' stato premuto il ritorno a capo
inline bool KBD_NL_READ(kbd_t *kbd)
{
	return kbd->nl_cnt > 0;
}

// indice nella keymap in base a flags
inline int KBD_SC_IDX(int flags)
{
	return (flags)&KBD_FLAG_SHIFT ?
		KBD_MOD_SHIFT: KBD_MOD_NONE;
}

// inserimento del codice nella coda (di caratteri) della tastiera attiva
//
static void kbd_put(unsigned short code)
{
	char ch;

	if(!kbd_att)
   return;

 ch = (char)(code&0xff);
 switch(ch) {
  case '\b':
   if(kbd_att->in != kbd_att->out &&
     kbd_att->queue[kbd_att->in > 0 ?
          kbd_att->in - 1: KBD_QUEUE_SIZE - 1] != '\n') {
    --kbd_att->in;
    kbd_att->in = kbd_att->in >= 0 ?
     kbd_att->in: KBD_QUEUE_SIZE - 1;
   }

   break;
  default:
   if(!ch)
    return;

   kbd_att->queue[kbd_att->in] = ch;
   if(ch == '\n')
    ++kbd_att->nl_cnt;
   kbd_att->in = (kbd_att->in + 1) % KBD_QUEUE_SIZE;
   if(kbd_att->in == kbd_att->out)
    kbd_att->out = (kbd_att->out + 1) %
     KBD_QUEUE_SIZE;
 }
}

// lettura di un carattere dalla coda di kbd
//
static int kbd_get(kbd_t *kbd)
{
 int rv;

 if(kbd->in == kbd->out)
  return -1;

 rv = kbd->queue[kbd->out];
 if(rv == '\n')
  --kbd->nl_cnt;
 kbd->out = (kbd->out + 1) % KBD_QUEUE_SIZE;

	return rv;
}

bool term_newchar(des_term *term, unsigned short code);

// processo esterno per la gestione dell' ingresso da tastiera
//
void tast_in(int h)
{
	char ch;
	unsigned short code;
	bool r;

	for(;;) {
		con_read(ch, r);

		if(r && kbd_att) {
			code = kbd_keymap[KBD_SC_IDX(kbd_att->flags)][ch&0x7f];

			if(!(ch&0x80) && KBD_IS_MOD(code))
				kbd_att->flags |= kbd_modflag_tab[KBD_MOD(code)];
			else if((ch&0x80) && KBD_IS_MOD(code))
				kbd_att->flags &= ~kbd_modflag_tab[KBD_MOD(code)];
			else if(!(ch&0x80)) {
				if(term_newchar(kbd_att->term, code))
					kbd_put(code);
			}
		}

		nwfi(master);
	}
}

extern "C" void abilita_tastiera(void);

// Interruzione hardware della tastiera
//
const int KBD_IRQ = 1;

// inizializzazione
//
void kbd_init()
{
	short id;
	bool r;

	activate_pe(tast_in, 0, PRIO_ESTERN_BASE + IRQ_MAX - KBD_IRQ,
		LIV_SISTEMA, id, tastiera, r);
	if(!r)
		panic("Impossibile creare processo esterno della tastiera");

	// l' inizializzazione avviene ad interruzioni disabilitate, quindi
	//  non ci saranno richieste di interruzione accolte fino all' uscita
	//  da io_init
	//
	abilita_tastiera();
}

// lettura di n caratteri da kbd
//  (deve essere chiamata ad interruzioni disabilitate)
//
int kbd_read(kbd_t *kbd, char *buf, int n)
{
 int i = 0, ch;

 while(i < n && (ch = kbd_get(kbd)) != -1) {
  buf[i++] = (char)ch;
  if(ch == '\n')
   break;
 }

 return i;
}

// numero di caratteri disponibili nella coda di kbd
//
int kbd_avail(kbd_t *kbd)
{
 if(kbd->in >= kbd->out)
  return kbd->in - kbd->out;

 return KBD_QUEUE_SIZE - kbd->out + kbd->in;
}

// rende kbd la tastiera virtuale attiva
//
int kbd_activate(kbd_t *kbd)
{
 kbd_att->flags = KBD_FLAG_NONE;
 kbd_att = kbd;

 return 0;
}

////////////////////////////////////////////////////////////////////////////////
//                             I/O DA TERMINALE                               //
////////////////////////////////////////////////////////////////////////////////

// descrittore di terminale virtuale
//
struct des_term {
	int mutex;	// protezione da utilizzo concorrente
	int sincr;	// sincronizzazione per la lettura da tastiera
	bool waiting;	// un processo attende su sincr

	kbd_t kbd;	// tastiera virtuale
	con_status cstat; // stato della console per il video virtuale
};

const int N_TERM = 12;
int term_att = 0;

des_term term_virt[N_TERM];

void attiva_term(int t)
{
	con_save(&term_virt[term_att].cstat);
	term_att = t;
	kbd_activate(&term_virt[term_att].kbd);
	con_load(&term_virt[term_att].cstat);
}

bool term_newchar(des_term *term, unsigned short code)
{
	char ch;

	if(code >= KBD_F1 && code <= KBD_F12) {
		attiva_term(KBD_FNUM(code));
		return false;
	}

	if(code == KBD_CANC && (term->kbd.flags&KBD_FLAG_CTRL) &&
			(term->kbd.flags&KBD_FLAG_ALT))
		reboot();

	if(code == '\b' && kbd_avail(&term->kbd) == 0)
		return false;

	if(code < 0xff) {
		ch = code&0xff;
		con_write(&ch, 1);
	}

	if(term->waiting && code == '\n') {
		term->waiting = false;
		sem_signal(term->sincr);
	}

	return true;
}

// NOTA: non richiede che vetti sia nello spazio condiviso, dato che l' handler
//  dell' interruzione scrive in un buffer interno
extern "C" void c_term_read_n(int term, char vetti[], int &quanti)
{
	des_term *p_des;
	int disp, letti;

	if(term < 0 || term >= N_TERM ||
			!verifica_area(&quanti, sizeof(int), true) ||
			!verifica_area(vetti, quanti, true))
		return;

	p_des = &term_virt[term];
	sem_wait(p_des->mutex);
	quanti = quanti > KBD_QUEUE_SIZE ? KBD_QUEUE_SIZE: quanti;

	lock();
	disp = kbd_avail(&p_des->kbd);

	if(disp >= quanti - 1 || KBD_NL_READ(&p_des->kbd)) {
		// sono gia' disponibili abbastanza caratteri, o
  		//  e' stato premuto invio
		letti = kbd_read(&p_des->kbd, vetti, quanti - 1);
	} else {
		// e' necessario bloccare
		p_des->waiting = true;
		unlock();

		sem_wait(p_des->sincr);

		lock();
		disp = kbd_avail(&p_des->kbd);
		letti = kbd_read(&p_des->kbd, vetti, quanti - 1);
	}

	unlock();

	vetti[letti] = 0;
	quanti = letti + 1;

	sem_signal(p_des->mutex);
}

// NOTA: non richiede che vetti sia nello spazio condiviso, dato che l' IO non
//  e' ad interruzione di programma
extern "C" void c_term_write_n(int term, char vetti[], int quanti)
{
	des_term *p_des;

	if(term < 0 || term >= N_TERM /*|| !verifica_area(vetti, quanti, false)*/)
		return;

	p_des = &term_virt[term];
	sem_wait(p_des->mutex);

	lock();
	if(term_att == term)
		con_write(vetti, quanti);
	else
		con_update(&term_virt[term].cstat, vetti, quanti);
	unlock();

	sem_signal(p_des->mutex);
}

void term_init(void)
{
	des_term *p_des;
	bool r1, r2;

	kbd_init();

	for(int i = 0; i < N_TERM; ++i) {
		p_des = &term_virt[i];

		sem_ini(p_des->mutex, 1, r1);
		sem_ini(p_des->sincr, 0, r2);
		if(!r1 || !r2)
			panic("Impossibile inizializzare i semafori per l' I/O");
		p_des->waiting = false;

		kbd_init(&p_des->kbd, p_des);
		con_init(&p_des->cstat);
   	}
}

////////////////////////////////////////////////////////////////////////////////
//            		GESTIONE DEL MECCANISMO DI DMA               	      //
////////////////////////////////////////////////////////////////////////////////

#define READ_BLOCK 0x44		// 0b01000100 MODE = single(01),noautoinit(0),
				//  addr.inc(0),lettura(01),canale(xx)
#define WRITE_BLOCK 0x48	// 0b01001000 MODE = single(01),noautoinit(0),
				//  addr.inc(0),scrittura(10),canale(xx)

const int D=2;

struct dma_reg {		// Registri del Controllore DMA comuni a tutti
	ind_b iCMD_STS;		//  i canali
	ind_b iMODER;
	ind_b iREQREG;
	ind_b iMASK_REG;
	ind_b iMASK;
};

struct dmachan_reg {		// Registri del Controllore DMA specifici per
	ind_b iPAGE;		//  ogni canale
	ind_b iMAR;
	ind_b iTCR;
	ind_b iCLEAR_FF;

};

extern "C" dma_reg cdma[D];	// Controllori DMA 8237 master(4-7) e slave (0-3)

extern "C" void dma_init();

// Trasferimento dati in DMA
//

extern "C" void setup_addr_dma(ind_b p,ind_b m,ind_fisico i_f);

extern "C" void setup_cont_dma(ind_b c,int c);

extern "C" void go_moder_in(unsigned char m,ind_b c,ind_b msk,short int ch);

extern "C" void go_moder_out(unsigned char m,ind_b c,ind_b msk,short int ch);

extern "C" void halt_moder(ind_b c,ind_b msk,short int ch);
	
inline void dma_status(ind_b reg,char &s) {
	inputb(reg,s);
}

////////////////////////////////////////////////////////////////////////////////
//               		GESTIONE DEL FLOPPY DISK               	      //
////////////////////////////////////////////////////////////////////////////////

// Codici degli errori comuni alle unita di memorizzazione di massa

#define D_ERR_NONE 0x00
#define D_ERR_BOUNDS 0xFF
#define D_ERR_PRESENCE 0xFE
// Specifici per il floppy
#define FD_ERR_NORECAL 0xFD
#define FD_ERR_NOSEEK 0xFC
#define FD_ERR_NOADDR 0xFB
#define FD_ERR_ADDRDMA 0xFA
#define FD_ERR_NOTX 0xF9

#define BLK_SIZE 512			// Dimensione del blocco (HD e FD)
#define H_BLK_SIZE (BLK_SIZE / 2)	// Meta' dimensione blocco (short int)

#define D_TIMEOUT 0x7FFF		// Timeout per l'accesso ai dischi
#define FD_RECAL_TIMEOUT 0xFF		// Timeout per il recalibrate del floppy

#define FD_STS_SIZE 7
#define FD_N_BLK 2880

// Geometria del floppy (fissa, floppy 1.44MB)
#define FD_TEST 2
#define FD_TR 80
#define FD_SETT_T 18
#define FD_MAX_BLK 128

// Operazione di lettura o scrittura
#define FDC_WRITE false
#define FDC_READ true
#define FD_RETRY 3  // Numero di tentadivi del driver del floppy

enum fd_cmd {   // Comandi per il controllore del floppy
 SPECIFY=0x03,
 RECALIBRATE=0x07,
 SEEK=0x0F,
 SENSEI=0x08,
 READ=0xE6,
 WRITE=0xC5,
 FORMAT=0x4D
};
 
struct fdcontr_reg {  // Registri del contrlollore del floppy
 ind_b iDOR;
 union {
  ind_b iM_STS;
  ind_b iD_RATE;
 }; //iM_STS_D_RATE
 ind_b iDATA;
 union {
  ind_b iDIR;
  ind_b iCCR;
 }; //iDIR_CCR
};

struct fddrive_data {  // Informazioni specifiche di un floppy drive
 bool motore;
 bool presente;
 char errore;
 int traccia;
};

struct des_fd {   // Descrittore di operazione di I/O su floppy
 fdcontr_reg indreg;
 dmachan_reg dmareg;
 fddrive_data drive[2];  // Gestiti fino a 2 drive
 fd_cmd curr_cmd;
 int curr_drv;
 int mutex;
 int sincr;
};

extern "C" des_fd fd;		// gestito un solo controllore

//Funzioni di utilita' per la gestione del disco
//

extern "C" int test_fd();

extern "C" void abilita_fd();

inline void fdc_delay() {
	char c;
	inputb((char*)0x80,c);
}

// Converte un blocco in testina-traccia-settore
//
void blk_tts(int primo, int &test, int &tr, int &sett) {
		//inversa di: primo=(2*FD_TR+FD_TEST)*FD_SETT_T+sett-1)
		test=(primo%(FD_SETT_T*FD_TEST))/(FD_SETT_T);
		tr=primo/(FD_SETT_T*FD_TEST);
		sett=(primo%FD_SETT_T)+1;
}

// Lettura dei blocchi non ancora trasferiti dal dma
// 
inline void get_quanti(des_fd *p_des,unsigned char &q) {
	char contL,contH;
	outputb(0x00,p_des->dmareg.iCLEAR_FF);
	inputb(p_des->dmareg.iTCR,contL);
	inputb(p_des->dmareg.iTCR,contH);
	q=(contH<<8)+(contL+1);
}

// Uscita di un byte verso il controllore floppy
//  la procedura richiede una temporizzazione precisa!
//
void outputb_fd(char a) {
	char stato;
	des_fd *p_des=&fd;
	short timeout=D_TIMEOUT;
	do {
		fdc_delay();			// Ritardo!
		inputb(p_des->indreg.iM_STS,stato);
		timeout--;
		if (timeout<0)
			panic("OUTPUT: Floppy disk timeout!\n");
	} while ((stato&0xC0)!=0x80);
	outputb(a,p_des->indreg.iDATA);
}

// Ingresso di un byte dal controllore floppy
//  la procedura richiede una temporizzazione precisa!
//
void inputb_fd(char &a) {
	char stato;
	des_fd *p_des=&fd;
	
	short timeout=D_TIMEOUT;
	do {
		fdc_delay();			// Ritardo!
		inputb(p_des->indreg.iM_STS,stato);
		timeout--;
		if (timeout<0)
			panic("INPUT: Floppy disk timeout!\n");
	} while ((stato&0xD0)!=0xD0);
	inputb(p_des->indreg.iDATA,a);
}

extern "C" void motore_on(ind_b c,bool &m,int d);

extern "C" void motore_off(ind_b c,bool &m);

// Invio di un comando (ultimo byte) al controllore e aggiornamento del
//  descrittore dell'operazione
//  
inline void invia_cmd_fd(des_fd *p_des,fd_cmd com) {
 outputb_fd(com);
 p_des->curr_cmd=com;
}

// Lettura dello stato del controllore floppy; da compiere in risposta
//  ad alcuni comandi
//  
void fdc_status(char buff[FD_STS_SIZE]) {
 char stato;
 des_fd *p_des=&fd;
 
 for (int i=0;i<FD_STS_SIZE;i++) {
  inputb(p_des->indreg.iM_STS,stato);
  if (!(stato&&0x10))
    break;
  inputb_fd(buff[i]);
 }
}

// Reset del piedino di interrupt del floppy e lettura dello stato del
//  dispositivo; da compiere in risposta ad alcuni comandi
//  
char fdc_sensei(int drv) {
 char stato;
 char curr_tr;
 des_fd *p_des=&fd;
 
 invia_cmd_fd(p_des,SENSEI);
 inputb_fd(stato);
 inputb_fd(curr_tr);
 if (drv!=-1)
  p_des->drive[drv].traccia=curr_tr;
 return stato;
}

// Riporta la testina sulla traccia 0; da compiere dopo il verificarsi di un
//  qualunque errore
//  
bool fdc_recalibrate(int drv,int test) {	// Ha un valore di ritorno 
	char stato;				//  potendo a sua volta fallire
	int timeout=0xFF;
	des_fd *p_des=&fd;
	
	if (!p_des->drive[0].motore)
		motore_on(p_des->indreg.iDOR,p_des->drive[drv].motore,drv);
	do {
		p_des->curr_drv=drv;
		invia_cmd_fd(p_des,RECALIBRATE);// quando e' possibile il floppy
		outputb_fd((test<<2)|drv);	//  deve funzionare ad
		sem_wait(p_des->sincr);		//  interruzione
		stato=p_des->drive[drv].errore;
	} while ((timeout--)>0 && (stato&0x30)==0x30);
	if ((stato&0xF8)!=0x20) {
		motore_off(p_des->indreg.iDOR,p_des->drive[drv].motore);
		p_des->drive[drv].errore=FD_ERR_NORECAL;
		return false;
	}
	p_des->drive[drv].errore=D_ERR_NONE;
	return true;
}

// Posizionamento della testina di lettura/scrittura; in molti floppy controller
//  non e' automatico come per gli hard disk
//  
bool fdc_seek(int drv,int test,int tr) {	// Prevede il ritorno anziche
	char stato;				//  bloccare il sistema
	des_fd *p_des=&fd;
	
	if (tr!=(p_des->drive[drv].traccia)) {
		p_des->curr_drv=drv;
		invia_cmd_fd(p_des,SEEK);
		outputb_fd((test<<2)|drv);
		outputb_fd(tr);
		sem_wait(p_des->sincr);		// Attesa dell'interruzione
		stato=p_des->drive[drv].errore;
		if ((stato&0xF8)!=0x20 || (p_des->drive[drv].traccia!=tr)) {
			motore_off(p_des->indreg.iDOR,p_des->drive[drv].motore);
			p_des->drive[drv].errore=FD_ERR_NOSEEK;
			return false;
		}
	}
	p_des->drive[drv].errore=D_ERR_NONE;
	return true;
}

// Impartisce un comando di lettura o scrittura al controllore; la funzione
//  e' unica essendo la struttura (peraltro complessa) dei due comandi molto
//  simile, almeno fino all'ultimo byte da inviare
//  
void fdc_read_write(int drv,bool read,int test,int tr,int sett) {
	char comando[8];
	des_fd *p_des=&fd;
 
 // Invio del comando di lettura o di scrittura
 if (read)
  invia_cmd_fd(p_des,READ);
 else
  invia_cmd_fd(p_des,WRITE);
 comando[0]=(test<<2)|drv; // Scelta del drive
 comando[1]=tr;   // Indirizzamento
 comando[2]=test;
 comando[3]=sett;
 comando[4]=2;   // 512 byte per settore
 comando[5]=18;   // Settori per traccia -> buffer FIFO
 comando[6]=0x1B;  // disco da 1.44 MB
 comando[7]=0xFF;  // DTL non utilizzato
 p_des->curr_drv=drv;
 for (int i=0;i<8;i++)
  outputb_fd(comando[i]);
}

// Avvio delle operazioni di lettura
//  (NB: puo' anche questa fallire)
//  
int startfd_in(des_fd *p_des,dma_reg *d_des,short drv,ind_w vetti,unsigned int primo,unsigned char &quanti) {
 char stato;
 ind_fisico iff;
 int test=0,tr=0,sett=1;   
 blk_tts(primo,test,tr,sett);  // Traduzione indirizz
 //Attivazione del motorino
 motore_on(p_des->indreg.iDOR,p_des->drive[drv].motore,drv); 
 outputb(0x00,p_des->indreg.iCCR); // DataRate 500Kb/sec
 if (!fdc_seek(drv,test,tr)) {  // La seek puo' fallire
  //sem_signal(p_des->sincr); //  (processo esterno)
  return -1;
 }

 trasforma((ind_l)vetti,iff);  // Ottenimento dell'indirizzo
      //  fisico (DMA!!)
 if (iff==0) {
  p_des->drive[drv].errore=FD_ERR_ADDRDMA;
  return -1;
 }

	// Inizializzazione del controllore DMA (indirizzi e contatore)
	setup_addr_dma(p_des->dmareg.iPAGE,p_des->dmareg.iMAR,iff);
	setup_cont_dma(p_des->dmareg.iTCR,BLK_SIZE*quanti);
	go_moder_in(READ_BLOCK,d_des->iMODER,d_des->iMASK,2);	// Programmazione

	fdc_read_write(drv,FDC_READ,test,tr,sett);		// Comando di lettura

	return test;		// ritorna la testina, ad uso della primitiva
}

// Avvio delle operazioni di scrittura
//  (NB: anche questa puo' fallire)
//  
int startfd_out(des_fd *p_des,dma_reg *d_des,short drv,ind_w vetti,unsigned int primo,unsigned char &quanti) {
	char stato;
	ind_fisico iff;
	int test=0,tr=0,sett=1;
	
	blk_tts(primo,test,tr,sett);		// Traduzione indirizzo
	//Attivazione del motorino
	motore_on(p_des->indreg.iDOR,p_des->drive[drv].motore,drv);
	outputb(0x00,p_des->indreg.iCCR);	// DataRate 500Kb/sec
	if (!fdc_seek(drv,test,tr)) {		// La seek puo' fallire
		//sem_signal(p_des->sincr);	//  (processo esterno)
		return -1;
	}
	
	trasforma((ind_l)vetti,iff);		// Ottenimento dell'indirizzo
						//  fisico (DMA!)
	if (iff==0) {
		p_des->drive[drv].errore=FD_ERR_ADDRDMA;
		return -1;
	}

	// Inizializzazione del controllore DMA (indirizzi e contatore)
	setup_addr_dma(p_des->dmareg.iPAGE,p_des->dmareg.iMAR,iff);
	setup_cont_dma(p_des->dmareg.iTCR,BLK_SIZE*quanti);
	go_moder_out(WRITE_BLOCK,d_des->iMODER,d_des->iMASK,2);	// Programmazione
	
	fdc_read_write(drv,FDC_WRITE,test,tr,sett);	// Comando di scrittura

	return test;		// ritorna la testina, ad uso della primitiva

}

// Primitiva di lettura dal floppy: legge quanti blocchi (max 128)
//  a partire da primo e li deposita in vetti (attraverso l'ind. fisico);
//  legge dal drive drv e riporta un fallimento in errore
//  
extern "C" void c_readfd_n(short drv,short vetti[],unsigned int primo,unsigned char &quanti,char &errore)
{
	des_fd *p_des;
	dma_reg *d_des;

	p_des=&fd;
	d_des=&cdma[0];		// Floppy = Controllore DMA Slave

	// Controllo di protezione
	if (drv<0 || drv>=2)
		return;
	if (!verifica_area(&quanti,sizeof(int),true) ||
	    !verifica_area(vetti,quanti,true) ||
	    !verifica_area(&errore,sizeof(char),true))
		return;

	// Controllo sulla selezione di un drive presente
	if (!p_des->drive[drv].presente) {
		errore=D_ERR_PRESENCE;
		return;
	}
	
	// Controllo sull'indirizzamento
	if (primo+quanti>FD_N_BLK || quanti>FD_MAX_BLK) {
		errore=D_ERR_BOUNDS;
		return;
	}

	sem_wait(p_des->mutex);

	for (int i=0;i<=FD_RETRY;i++) {		// Tenta la lettura piu' volte
	p_des->drive[drv].errore=D_ERR_NONE;	// Reset errori precedenti
	// Avvio operazioni di lettura (la chiamata puo' fallire!)
	int test=startfd_in(p_des,d_des,drv,vetti,primo,quanti);
	if (test!=-1) {				// Se qualcosa e' andato storto
		sem_wait(p_des->sincr);		//  non ci blocchiamo
		if (p_des->drive[drv].errore==D_ERR_NONE)
			break;
		else
			if (i!=FD_RETRY)
				fdc_recalibrate(drv,test); // prima di riprovare
		}					
	}

	get_quanti(p_des,quanti);
	errore=p_des->drive[drv].errore;
	sem_signal(p_des->mutex);

}

// Primitiva di scrittura su floppy: scrive quanti blocchi (max 128)
//  a partire da primo prelevandoli da vetti (attraverso l'ind. fisico);
//  scrive sul drive drv e riporta un fallimento in errore
//  
extern "C" void c_writefd_n(short drv,ind_w vetti,unsigned int primo,unsigned char &quanti,char &errore)
{
	des_fd *p_des;
	dma_reg *d_des;

	p_des=&fd;
 	d_des=&cdma[0];  // Floppy = Controllore DMA Slave

 	// Controllo di protezione
 	if (drv<0 || drv>=2)
  		return;
 	if (!verifica_area(&quanti,sizeof(int),true) ||
     !verifica_area(vetti,quanti,true) ||
     !verifica_area(&errore,sizeof(char),true))
  return;

 //Controllo sulla selezione di un drive presente
 if (!p_des->drive[drv].presente) {
  errore=D_ERR_PRESENCE;
  return;
 }
 
 //Controllo sull'indirizzamento
 if (primo+quanti>FD_N_BLK || quanti>FD_MAX_BLK) {
  errore=D_ERR_BOUNDS;
  return;
 }

 sem_wait(p_des->mutex);
 
 for (int i=0;i<=FD_RETRY;i++) {  // Tenta la scrittura piu' volte
 p_des->drive[drv].errore=D_ERR_NONE; // reset errori precedenti
 // Avvio operazioni di lettura (la chiamata puo' fallire!)
 int test=startfd_out(p_des,d_des,drv,vetti,primo,quanti);
 if (test!=-1) {    // Se qualcosa e' andato storto
  sem_wait(p_des->sincr);  //  non ci blocchiamo
  if (p_des->drive[drv].errore==D_ERR_NONE)
   break;
  else
   if (i!=FD_RETRY)
    fdc_recalibrate(drv,test); // prima di riprovare
  }
 }
 
	get_quanti(p_des,quanti);
	errore=p_des->drive[drv].errore;
	sem_signal(p_des->mutex);
}

// Processo esterno per il floppy disk (in DMA)
// 
void dmaest_fd(int h) {
	char stato[FD_STS_SIZE];
	char s_dma;
	int drv;
	des_fd *p_des=&fd;
	dma_reg *d_des=&cdma[0];
	
	for (;;) {
		drv=p_des->curr_drv;	// C'e' la possibilita' che arrivino
		p_des->curr_drv=-1;	//  interrupt non richiesti
		if (drv!=-1) {
			if (p_des->curr_cmd!=READ && p_des->curr_cmd!=WRITE) {
			// INT prodotte da comandi di seek o recalibrate
					p_des->drive[drv].errore=fdc_sensei(drv);
			}
   else {
    fdc_status(stato); // Risposta corretta
    motore_off(p_des->indreg.iDOR,p_des->drive[drv].motore);
    halt_moder(d_des->iREQREG,d_des->iMASK,2);
    if ((stato[0]&0xC0)!=0x00)
     p_des->drive[drv].errore=FD_ERR_NOTX;
    else {
     p_des->drive[drv].errore=D_ERR_NONE;
     p_des->drive[drv].traccia=stato[3];
    }
    dma_status(d_des->iCMD_STS,s_dma);
   }
   sem_signal(p_des->sincr); // In ogni caso sveglia
  }     //  il processo
  else
   fdc_sensei(-1);  // Interruzione da ignorare
  nwfi(master);   // EOI al master PIC
 }
}
 
const int FDC_IRQ=6;

// Inizializzazione del contollore floppy; max 1 controllore e 2 drive da 1.44MB
// 
void fdc_init() {
 char stato;
 int n_drive;
 char buff[FD_STS_SIZE];

 des_fd *p_des;
 
 short id;
 bool r1,r2;
 
 p_des=&fd;

 p_des->drive[0].presente=true;
 n_drive=test_fd();  // Controlla la presenza di uno o piu'
 if (n_drive==1)   //  drive dalla CMOS
  p_des->drive[1].presente=false;
 else
  p_des->drive[1].presente=true;
	outputb(0x00,p_des->indreg.iDOR);	// Disabilita il controllore
	outputb(0x00,p_des->indreg.iD_RATE);	// Data Rate	
	outputb(0x0C,p_des->indreg.iDOR);	// Riabilita il controllore
	p_des->drive[0].motore=false;		//  i motori sono spenti
	p_des->drive[1].motore=false;
	invia_cmd_fd(p_des,SPECIFY);		// Imposta i parametri del floppy
	outputb_fd(0xDF);			// parametri secondari
	outputb_fd(0x06);			//  HeadLoadTime=6ms, DMA=si
	for (int d=0;d<n_drive && p_des->drive[d].presente;d++) {
		motore_on(p_des->indreg.iDOR,p_des->drive[d].motore,d);
		/*	RECALIBRATE	*/
		int i=0;		// Reimplementata per funzionare in
		do {			//  polling (int. sono disabilitate)
			invia_cmd_fd(p_des,RECALIBRATE);
			outputb_fd(d);
			stato=fdc_sensei(d);
			fdc_delay();	// Si rinuncia se non si riesce
		} while ((stato&0x30)==0x30 && (++i)<=FD_RECAL_TIMEOUT);
		/*	----------	*/
		motore_off(p_des->indreg.iDOR,p_des->drive[d].motore);
	}
	p_des->curr_drv=-1;

	sem_ini(p_des->mutex, 1, r1);
	sem_ini(p_des->sincr, 0, r2);
	if(!r1 || !r2)
		panic("\nImpossibile allocare i semafori per l' IO");
	activate_pe(dmaest_fd,0,PRIO_ESTERN_BASE+IRQ_MAX-FDC_IRQ,LIV_SISTEMA,id,floppy,r1);
 if (!r1 || !r2)
  panic("\nImpossibile creare un processo esterno del floppy disk");
 abilita_fd();  // Abilita le interruzioni dal floppy
}

////////////////////////////////////////////////////////////////////////////////
//                 GESTIONE DELL'HARD DISK                      //
////////////////////////////////////////////////////////////////////////////////

#define STRICT_ATA

#define HD_MAX_TX 256

#define HD_STS_ERR 0x01  //Maschere per i biti fondamentali dei registri
#define HD_STS_DRQ 0x08
#define HD_STS_DRDY 0x40
#define HD_STS_BSY 0x80
#define HD_DRV_MASTER 0
#define HD_DRV_SLAVE 1

const int A=2;

enum hd_cmd {   // Comandi per i controllori degli hard disk
 NONE=0x00,
 IDENTIFY=0xEC,
 SET_FEATURES=0xEF,
 SET_MULT_MODE=0xC6,
 WRITE_SECT=0x30,
 READ_SECT=0x20,
 HD_SEEK=0x70,
 DIAGN=0x90
};

struct interfata_reg {
 union {    // Command Block Register
  ind_b iCMD;
  ind_b iSTS;
 }; //iCMD_iSTS;
 ind_w iDATA;
 union {
  ind_b iFEATURES;
  ind_b iERROR;
	}; //iFEATURES_iERROR
	ind_b iSTCOUNT;
	ind_b iSECT_N;
	ind_b iCYL_L_N;
	ind_b iCYL_H_N;
	union {
		ind_b iHD_N;
		ind_b iDRV_HD;
	}; //iHD_N_iDRV_HD
	union {				// Control Block Register
		ind_b iALT_STS;		
		ind_b iDEV_CTRL;
	}; //iALT_STS_iDEV_CTRL
};

struct geometria {			// Struttura 3-D di un disco
	unsigned short cil;	
	unsigned short test;
	unsigned short sett;
	unsigned int tot_sett;		// Utile in LBA
};

struct drive {				// Drive su un canale ATA
	bool presente;
	bool dma;
	geometria geom;
};

struct des_ata {		// Descrittore operazione per i controller ATA
	interfata_reg indreg;
	drive disco[2];		// Ogni controller ATA gestisce 2 drive
	hd_cmd comando;
	char errore;
	char cont;
	ind_w punt;	
	int mutex;
	int sincr;
};

extern "C" des_ata hd[A];	// 2 controller ATA

const estern_id hd_id[2]={ata0,ata1};

//Funzioni di utilita' per la gestione del disco
//
inline bool hd_drq(char &s) {		// Test di Data ReQuest
	return s&HD_STS_DRQ;
}

inline bool hd_errore(char &s) {	// Test di un eventuale errore (generico)
	return s&HD_STS_ERR;
}

inline bool hd_drdy(char &s) {		// Test di Drive ReaDY
	return s&HD_STS_DRDY;
}

// Attende una particolare configurazione di bit nel registro di stato del
//  controller; tipicamente deve resettarsi BuSY
//  
char aspetta_STS(des_ata *p_des,char bit,bool &ris) {
	char stato,test;
	short timeout=D_TIMEOUT;		// Previsto un timeout
	do {
		inputb(p_des->indreg.iALT_STS,stato);
		if (bit==HD_STS_DRDY || bit==HD_STS_DRQ)
			test=(~stato);
		else
			test=stato;
		test&=bit;
		timeout--;
		if (timeout<0) {
			ris=false;		// Scaduto il timeout!
			return stato;
		}
	} while (test!=0);
	ris=true;
	return stato;
}

// Invia un comando al controller con la temporizzazione corretta
// 
inline bool invia_cmd(des_ata *p_des,hd_cmd com) {
#ifdef STRICT_ATA
	bool ris=true;
	aspetta_STS(p_des,HD_STS_DRDY,ris);
	if (!ris)
		return false;
#endif
	p_des->comando=com;
	outputb(com,p_des->indreg.iCMD);
	return true;
}

// Legge il buffer di uscita del controller se questo contiene dati
// 
void leggi_sett_buff(des_ata *p_des,short vetti[],char &stato) {
 if (!hd_drq(stato))  
  // DRQ _deve_ essere settato!
  panic("\nErrore Grave: DRQ non e' settato!");
 inputbuffw(p_des->indreg.iDATA,vetti,H_BLK_SIZE);
 //for (int i=0;i<H_BLK_SIZE;i++) //<=== Forma equivalente (debugging)
 // inputw(p_des->indreg.iDATA,vetti[i]);
}

// Scrive il buffer di ingresso del controller se questo aspetta dati
// 
void scrivi_sett_buff(des_ata *p_des,short vetti[],char &stato) {
 if (!hd_drq(stato))  
  // DRQ _deve_ essere settato!
  panic("\nErrore Grave: DRQ non e' settato!");
 outputbuffw(vetti,p_des->indreg.iDATA,H_BLK_SIZE);
 //for (int i=0;i<H_BLK_SIZE;i++) //<=== Forma equivalente (debugging)
 // outputw(vetti[i],p_des->indreg.iDATA);
}

//Funzioni per l'ingresso/uscita dati e processo estern0
//
extern "C" void go_inouthd(ind_b i_dev_ctl);

extern "C" void halt_inouthd(ind_b i_dev_ctl);

extern "C" bool test_canale(ind_b st,ind_b stc,ind_b stn);

extern "C" int leggi_signature(ind_b stc,ind_b stn,ind_b cyl,ind_b cyh);

extern "C" void abilita_hd(ind_b h);

extern "C" bool drive_reset(ind_b dvctrl,ind_b st);

extern "C" void set_drive(short ms,ind_b i_drv_hd);

extern "C" void get_drive(ind_b i_drv_hd,short &ms);

extern "C" void setup_addr_hd(des_ata *p,unsigned int primo);	

// Trasferimento dati in PIO Mode
//

// Avvio delle operazioni di lettura da hard dik
// 
void starthd_in(des_ata *p_des,short drv,short vetti[],unsigned int primo,unsigned char &quanti)	{
	p_des->punt=vetti;
#ifdef STRICT_ATA
	bool ris;
	aspetta_STS(p_des,HD_STS_BSY,ris);	// aspetta "drive not BuSY"
#endif
	set_drive(drv,p_des->indreg.iDRV_HD);	// Selezione del drive richiesto
#ifdef STRICT_ATA
	char stato=aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta di nuovo!!
	if (!hd_drdy(stato) || hd_errore(stato))
			panic("Errore nell'inizializzazione della lettura!");
#endif
	setup_addr_hd(p_des,primo); 
	outputb(quanti,p_des->indreg.iSTCOUNT);
	// Abilita le interruzioni per il controller-drive selezionato
	go_inouthd(p_des->indreg.iDEV_CTRL);	
	invia_cmd(p_des,READ_SECT);
}

// Avvio delle operazioni di scrittura su hard disk
//
void starthd_out(des_ata *p_des,short drv,short vetti[],unsigned int primo,unsigned char &quanti)	
{	char stato;				
	bool ris;
	p_des->punt=vetti;
#ifdef STRICT_ATA
	aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta "drive not BuSY"
#endif
	set_drive(drv,p_des->indreg.iDRV_HD);	// Selezione del drive richiesto
#ifdef STRICT_ATA
	stato=aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta di nuovo!
	if (!hd_drdy(stato) || hd_errore(stato))
			panic("Errore nell'inizializzazione della scrittura!");
#endif
	setup_addr_hd(p_des,primo); 
	outputb(quanti,p_des->indreg.iSTCOUNT);
 //Abilita le interruzioni per il controller-drive selezionato
 go_inouthd(p_des->indreg.iDEV_CTRL);  
 invia_cmd(p_des,WRITE_SECT); // Asimmetria con la lettura, bisogna
#ifdef STRICT_ATA   //  scrivere il primo blocco
 aspetta_STS(p_des,HD_STS_BSY,ris); // Aspetta "drive not BuSY"
#endif
 // Aspetta DRQ, previsto dal protocollo di comunicazione
 stato=aspetta_STS(p_des,HD_STS_DRQ,ris);
 lock();
 inputb(p_des->indreg.iSTS,stato);
 scrivi_sett_buff(p_des,p_des->punt,stato);
 p_des->punt+=H_BLK_SIZE;
 unlock();
}

// Primitiva di lettura da hard disk: legge quanti blocchi (max 256, char!)
//  a partire da primo depositandoli in vetti; il controller usato e ind_ata,
//  il drive e' drv. Riporta un fallimento in errore
//  
extern "C" void c_readhd_n(short ind_ata,short drv,short vetti[],unsigned int primo,unsigned char &quanti,char &errore)
{
 des_ata *p_des;

 p_des = &hd[ind_ata];

 // Controllo sulla protezione
 if (ind_ata<0 || ind_ata>=A || drv<0 || drv>=2)
  return;
 if (!verifica_area(&quanti,sizeof(int),true) || !verifica_area(vetti,quanti,true) || !verifica_area(&errore,sizeof(char),true))
  return;

 // Controllo sulla selezione di un drive presente
 if (p_des->disco[drv].presente==false) {
  errore=D_ERR_PRESENCE;
  return;
 }

 // Controllo sull'indirizzamento
 if (primo>p_des->disco[drv].geom.tot_sett)
  errore=D_ERR_BOUNDS;
 else {  
  sem_wait(p_des->mutex);
  p_des->errore=D_ERR_NONE;  // Reset degli errori precedenti 
  p_des->disco[drv].dma=false; // Trasferimento in PIO Mode
  // Abilitazione del controller all'operazione
  starthd_in(p_des,drv,vetti,primo,quanti);
  sem_wait(p_des->sincr);
  quanti=p_des->cont;
  errore=p_des->errore;
  sem_signal(p_des->mutex);
 }
}

// Primitiva di scrittura su hard disk: scrive quanti blocchi (max 256, char!)
//  a partire da primo prelevandoli da vetti; il controller usato e' ind_ata,
//  il drive e' drv. Riporta un fallimento in errore
//  
extern "C" void c_writehd_n(short ind_ata,short drv,short vetti[],unsigned int primo,unsigned char &quanti,char &errore)
{
 des_ata *p_des;

 p_des = &hd[ind_ata];

 // Controllo sulla protezione
 if (ind_ata<0 || ind_ata>=A || drv<0 || drv>=2)
  return;
 if (!verifica_area(&quanti,sizeof(int),true) || !verifica_area(vetti,quanti,true) || !verifica_area(&errore,sizeof(char),true))
  return;

 // Controllo sulla selezione di un drive presente
 if (p_des->disco[drv].presente==false) {
  errore=D_ERR_PRESENCE;
  return;
 }
 // Controllo sull'indirizzamento
 if (primo>p_des->disco[drv].geom.tot_sett)
  errore=D_ERR_BOUNDS;
 else {
  sem_wait(p_des->mutex);
  p_des->errore=D_ERR_NONE;  // Reset degli errori precedenti
  p_des->disco[drv].dma=false; //Trasferimento in PIO Mode
  starthd_out(p_des,drv,vetti,primo,quanti);
  sem_wait(p_des->sincr);
  quanti=p_des->cont;
  errore=p_des->errore;
  sem_signal(p_des->mutex);
 }
}

// Processo esterno per l'hard disk, valido sia per l'ingresso che per 
//  l'uscita
//
void estern_hd(int h)
{
 des_ata *p_des;
 char stato;
 short drv;
 hd_cmd curr_cmd;
 bool fine;

 p_des = &hd[h];  // In base al parametro si sceglie un controller

 for(;;) {
  curr_cmd=p_des->comando; // Gestione delle interruzioni
  p_des->comando=NONE;  //  non richieste
  if (curr_cmd!=NONE) {
   get_drive(p_des->indreg.iDRV_HD,drv);
#ifdef STRICT_ATA
   bool ris;
   stato=aspetta_STS(p_des,HD_STS_BSY,ris);
#endif /* STRICT_ATA */
   fine=false;
   inputb(p_des->indreg.iSTS,stato);
   if (!hd_errore(stato) && (curr_cmd==READ_SECT || curr_cmd==WRITE_SECT)) {
    if (curr_cmd==WRITE_SECT) {
     inputb(p_des->indreg.iSTCOUNT,p_des->cont);
     if (p_des->cont==0)
      fine=true;
     else {
      scrivi_sett_buff(p_des,p_des->punt,stato);
      p_des->comando=WRITE_SECT;
     }
    }
    else {  // Comando di READ_SECT
     leggi_sett_buff(p_des,p_des->punt,stato);
     inputb(p_des->indreg.iSTCOUNT,p_des->cont);
     if (p_des->cont==0)
      fine=true;
     else
      p_des->comando=READ_SECT;
    }
    p_des->punt+=H_BLK_SIZE;
   }
   else {
    inputb(p_des->indreg.iERROR,p_des->errore);
    fine=true;
   }
   if (fine==true) {
    //Disabilita le interruzioni dal controller
    halt_inouthd(p_des->indreg.iDEV_CTRL); 
				if (curr_cmd!=IDENTIFY && curr_cmd!=DIAGN)
					sem_signal(p_des->sincr);
			}	
		}
		nwfi(slave); 	// Controller ATA = Slave PIC (entrambi, 14 e 15)
	}
}

const int ATA1_IRQ=15;

// Inizializzazione e autoriconoscimento degli hard disk collegati ai canali ATA
// 
void hd_init() {
	char stato=0;
	bool test;
	int init_time=D_TIMEOUT;
	short st_sett[H_BLK_SIZE];
	short id;
	bool r,r1,r2;
	int hd_base_prio=PRIO_ESTERN_BASE+IRQ_MAX-ATA1_IRQ;
	short canali=0;
	
	des_ata *p_des;
	for (int i=0;i<A;i++) {				// Per ogni controller
		p_des=&hd[i];
		// Disabilita le interruzioni
		halt_inouthd(p_des->indreg.iDEV_CTRL);	
	}
	for (int i=0;i<A;i++) {			// Per ogni controller
  p_des=&hd[i];
  short drv=HD_DRV_MASTER;
  set_drive(drv,p_des->indreg.iDRV_HD); //LBA e drive drv
  // Procedura diversificata una volta effettuato il test()
  if (test_canale(p_des->indreg.iSTS,p_des->indreg.iSTCOUNT,p_des->indreg.iSECT_N)) {
#ifndef BOCHS  // In sistemi reali e' necessario inviare questo comando
   set_drive(HD_DRV_MASTER,p_des->indreg.iDRV_HD);
   invia_cmd(p_des,DIAGN);  // Diagnostica dei drive
   do
    inputb(p_des->indreg.iSTS,stato);
   while (stato&HD_STS_BSY); // Polling
   inputb(p_des->indreg.iERROR,stato);
   do
    inputb(p_des->indreg.iSTS,stato);
   while (!stato&HD_STS_DRDY); // Polling
#endif  // Diagnostica terminata

   for (int d=0;d<2;d++) {  // Per ogni drive
    set_drive(drv,p_des->indreg.iDRV_HD);
    p_des->disco[d].dma=false;
#ifdef BOCHS  // BOCHS ha bisogno del reset dei controller
    if (!drive_reset(p_des->indreg.iDEV_CTRL,p_des->indreg.iSTS)) {
     p_des->disco[d].presente=false;
     drv=HD_DRV_SLAVE;
     continue;
    }
    else
     p_des->disco[d].presente=true;
#endif  // A questo punto in entrambi i casi e' disponibile la signature
    if (leggi_signature(p_des->indreg.iSTCOUNT,p_des->indreg.iSECT_N,p_des->indreg.iCYL_L_N,p_des->indreg.iCYL_H_N)!=0) {
     p_des->disco[d].presente=false;
     drv=HD_DRV_SLAVE;
     continue;
    }
    else
     p_des->disco[d].presente=true;
  
  //Riconoscimento della geometria e dei parametri del disco
    if (!invia_cmd(p_des,IDENTIFY)) { 
     p_des->disco[d].presente=false;
					drv=HD_DRV_SLAVE;
					continue;
				}
#ifdef STRICT_ATA
				bool ris;
				aspetta_STS(p_des,HD_STS_BSY,ris);
#endif
				inputb(p_des->indreg.iSTS,stato);
				leggi_sett_buff(p_des,st_sett,stato);
				// Inizializzazione della geometria
				p_des->disco[d].geom.cil=st_sett[1];		
				p_des->disco[d].geom.test=st_sett[3];
				p_des->disco[d].geom.sett=st_sett[6];
				p_des->disco[d].geom.tot_sett=(st_sett[58]<<16)+st_sett[57];
				drv=HD_DRV_SLAVE;
			}
			sem_ini(p_des->mutex, 1, r1);
			sem_ini(p_des->sincr, 0, r2);
			if(!r1 || !r2)
				panic("\nImpossibile allocare i semafori per l' IO");
			activate_pe(estern_hd,i,hd_base_prio-i,LIV_SISTEMA,id,hd_id[i],r);
			if(!r)
				panic("\nImpossibile creare un processo esterno dell'hard disk");
		}
		else
			p_des->disco[0].presente=p_des->disco[1].presente=false;
		p_des->comando=NONE;
		abilita_hd(p_des->indreg.iSTS);
	}
}

// Funzione di utilita' perche' l'utente possa conoscere se interessato,
//  le dimensioni e la struttura dei suoi hard disk
//  
extern "C" void c_geometria(int interf,short drv,unsigned short &c,unsigned short &t,unsigned short &s,int &ts,char &err) {
 if (hd[interf].disco[drv].presente==true) {
  c=hd[interf].disco[drv].geom.cil;
  t=hd[interf].disco[drv].geom.test;
  s=hd[interf].disco[drv].geom.sett;
  ts=hd[interf].disco[drv].geom.tot_sett;
  err=D_ERR_NONE;
 }
 else
  err=D_ERR_PRESENCE;
}

//////////////////////////////////////////////////////////////////////////////
//       SUPPORTO AI RIMPIAZZAMENTI DELLE PAGINE IN MEMORIA DI MASSA        //
//////////////////////////////////////////////////////////////////////////////

#define PAGE_SIZE 4096

// Se out=true: sposta QUANTE pagine di memoria a partire da quella puntata
//  da PAG (come short*) sul disco DRV del controller IND_ATA a partire dal
//  PRIMO blocco
// Se out=false: sposta QUANTE pagine di memoria alla locazione puntata
//  da PAG (come short*) prelevandole dal disco DRV del controller IND_ATA
//  in posizione PRIMO 
//  
extern "C" void c_swap_pag(bool out,short ind_ata,short drv,unsigned int primo,void* pag,int quante) {
 des_ata *p_des;
 bool ris;
 char stato;

 p_des = &hd[ind_ata];

 p_des->punt=(short*)pag;
#ifdef STRICT_ATA
 aspetta_STS(p_des,HD_STS_BSY,ris); // aspetta "drive not BuSY"
#endif
	set_drive(drv,p_des->indreg.iDRV_HD);	// Selezione del drive richiesto
#ifdef STRICT_ATA
	stato=aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta di nuovo!!
	if (!hd_drdy(stato) || hd_errore(stato))
			panic("Errore durante un rimpiazzamento");
#endif
	// Disabilita le interruzioni per il controller/drive selezionato
	halt_inouthd(p_des->indreg.iDEV_CTRL);

	quante=(quante*PAGE_SIZE)/BLK_SIZE;
	do {
		setup_addr_hd(p_des,primo);
		outputb((quante<HD_MAX_TX?quante:0),p_des->indreg.iSTCOUNT);
		quante-=HD_MAX_TX;
		if (!out)
			invia_cmd(p_des,READ_SECT);
		else
			invia_cmd(p_des,WRITE_SECT);
		do {		
#ifdef STRICT_ATA
			aspetta_STS(p_des,HD_STS_BSY,ris);	// Aspetta "drive not BuSY"
#endif
			// Aspetta DRQ, previsto dal protocollo di comunicazione
			stato=aspetta_STS(p_des,HD_STS_DRQ,ris);
			inputb(p_des->indreg.iSTS,stato);
			if (!out)
				leggi_sett_buff(p_des,p_des->punt,stato);
			else
				scrivi_sett_buff(p_des,p_des->punt,stato);	
			inputb(p_des->indreg.iSTS,stato);
			if (hd_errore(stato))
				panic("Errore durante un rimpiazzamento");
			p_des->punt+=H_BLK_SIZE;
			inputb(p_des->indreg.iSTCOUNT,p_des->cont);
		} while ((p_des->cont)!=0);
		primo+=HD_MAX_TX;
 } while (quante>0);
}






////////////////////////////////////////////////////////////////////////////////
//				Funzioni di utilit			      //
////////////////////////////////////////////////////////////////////////////////

// Replicata in sistema.cpp
//


void *memcpy(void *dest, const void *src, unsigned int n)
{
	char *dest_ptr = (char *)dest, *src_ptr = (char *)src;
	int i;

	for(i = 0; i < n; ++i)
		*dest_ptr++ = *src_ptr++;

	return dest;
}

// Replicata in sistema.cpp
//
void *memset(void *dest, int c, size_t n)
{
        size_t i;

        for(i = 0; i < n; ++i)
              ((char *)dest)[i] = (char)c;

        return dest;
}

//funzioni di libreria per le stringhe


int strlen(const char *s)
{
	int l = 0;

	while(*s++)
		++l;

	return l;
}

char *strncpy(char *dest, const char *src, size_t l)
{
	size_t i;

	for(i = 0; i < l && src[i]; ++i)
		dest[i] = src[i];

	return dest;
}

static const char hex_map[16] = { '0', '1', '2', '3', '4', '5', '6', '7',
	'8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };

static void htostr(char *buf, unsigned long l)
{
	int i;

	buf[0] = '0';
	buf[1] = 'x';

	for(i = 9; i > 1; --i) {
		buf[i] = hex_map[l % 16];
		l /= 16;
	}
}

static void itostr(char *buf, unsigned int len, long l)
{
	int i, div = 1000000000, v, w = 0;

	if(l == (-2147483647 - 1)) {
		strncpy(buf, "-2147483648", 12);
		return;
	} else if(l < 0) {
		buf[0] = '-';
		l = -l;
		i = 1;
	} else if(l == 0) {
		buf[0] = '0';
		buf[1] = 0;
		return;
	} else
		i = 0;

	while(i < len - 1 && div != 0) {
		if((v = l / div) || w) {
			buf[i++] = '0' + (char)v;
			w = 1;
		}

		l %= div;
		div /= 10;
	}

	buf[i] = 0;
}

//type def
typedef unsigned char u8; //intero unsigned su 8 bit
typedef unsigned short int  u16; //intero unsigned su 16 bit
typedef unsigned int u32; //intero unsigned su 32 bit
typedef void* punt;//puntatore generico

//funzione di utilit che serve per creare in maniera dinamica strutture di n byte allineate
//al byte alg
extern "C" void* mem_alloc_alg(u32 byte,u32 alg)
{
	lock();//operazione da fare in mutua esclusione
	bool stato=true;
	void* punt;
	void* stru;
	for(int i=0;stato==true;i++)
	{
		punt=mem_alloc(i);
		stru=mem_alloc(byte);
		ind_fisico ind;
		trasforma((ind_l)stru,ind);
		if((ind % alg)==0)
			stato=false;
		else
		{
			mem_free(punt);
			mem_free(stru);
		}
	}
	mem_free(punt);
	unlock();
	return stru;
}


//funzione usata per stampare sul terminale di log dello spazio di I/O
void dbg(char* str)
{
	int lun=strlen(str);
	c_term_write_n(10,str,lun);
}

////////////////////////////////////////////////////////////////////////////////
//				Driver PCI				      //
////////////////////////////////////////////////////////////////////////////////

//Questo  un driver PCI molto semplificato costruito solo per ricercare 
//un tipo di periferica


//funzione usata per interrogare il bus pci
int pci_interr(int bus,int dev,int fun,int dw){
	int reg=0;
	int ris=0;
	reg |= (1 << 31);//abilito l'interrogazione
	reg |= (dw << 2);//double word da leggere
	reg |= (fun << 8);//funzione
	reg |= (dev << 11);//dev number
	reg |= (bus << 16);//bus number	
	outputl(reg, (ind_b)0x80000cf8);
	inputl((ind_b)0x80000cfc,ris);
	return ris;
}

#define class_code(num)		(num >> 24)
#define sub_class_code(num)	((num >> 16) & 0xff)
#define prog_if(num)		((num >> 8) & 0xff)

#define MAX_PCI 255
#define MAX_DEV 31
#define MAX_FUN 7
#define MAX_DW 63

#define USB_CLASS_CODE		0x0c
#define USB_SUB_CLASS_CODE	0x03
#define USB_PROG_IF		0x00

//Questa funzione ricerca un device nello spazio pci in base al suo class_code e sub_class_code e prog_if
//si ferma alla prima occorrenza.Restituisce 1 se ha trovato il device in question altrimenti 0

int pci_find(int class_code,int sub_class_code,int prog_if,int& bus,int& dev,int& fun)
{
	for(int i=0;i<MAX_PCI;i++)
		for(int j=0;j<MAX_DEV;j++)
			for(int h=0;h<MAX_FUN;h++)
			{
				int ris=pci_interr(i,j,h,2);
				if((class_code(ris)==class_code) && (sub_class_code(ris)==sub_class_code) && (prog_if(ris)==prog_if))
				{
					bus=i;
					dev=j;
					fun=h;
					return true;
				}
			}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//                 Driver UHCI (Universal Host Controller)                    //
////////////////////////////////////////////////////////////////////////////////

#define DEBUG_USB
//#undef DEBUG_USB

//viene definito un timeout per le operazioni a controllo di programma espresso in cicli di clock
#define TIMEOUT		1000000000

#define NUM_FRAME_POINTER 1024

//alloca i TD
#define alloc_td()	(td_des*)mem_alloc_alg(sizeof(td_des),16)
#define alloc_qh()	(qh_des*)mem_alloc_alg(sizeof(qh_des),16)
#define alloc_uhci_framelist()		(uhci_framelist*)mem_alloc_alg(sizeof(uhci_framelist),4096)

//limitazioni del driver:
//di tutte le porte USB ne individua una sola e la inizializza e usa solo quella
//al bus usb pu essere collegata una sola periferica alla volta
//non viene gestito lo scollegamento



//strutture dati per la rappresentazione di un USB device generico

#define USB_MAXCONFIG		1
#define USB_MAXINTERFACES	3
#define USB_MAXENDPOINTS	3

//descrittore di un device generico
struct device_des {
	u8  bLength;
	u8  bDescriptorType;
	u16 bcdUSB;
	u8  bDeviceClass;
	u8  bDeviceSubClass;
	u8  bDeviceProtocol;
	u8  bMaxPacketSize0;
	u16 idVendor;
	u16 idProduct;
	u16 bcdDevice;
	u8  iManufacturer;
	u8  iProduct;
	u8  iSerialNumber;
	u8  bNumConfigurations;
};

//descrittore di un endpoint
struct endpoint_des {
	u8  bLength;
	u8  bDescriptorType;
	u8  bEndpointAddress;
	u8  bmAttributes;
	u16 wMaxPacketSize;
	u8  bInterval;
};

//descrittore di un interfaccia
struct interface_des {
	//u8  bLength;
	u8  bDescriptorType;
	u8  bInterfaceNumber;
	u8  bAlternateSetting;
	u8  bNumEndpoints;
	u8  bInterfaceClass;
	u8  bInterfaceSubClass;
	u8  bInterfaceProtocol;
	u8  iInterface;
	endpoint_des endpoint[USB_MAXENDPOINTS];
};

//descrittore di una configurazione
struct config_des {
	u8  bLength;
	u8  bDescriptorType;
	u16 wTotalLength;
	u8  bNumInterfaces;
	u8  bConfigurationValue;
	u8  iConfiguration;
	u8  bmAttributes;
	u8  MaxPower;
	interface_des interface[USB_MAXINTERFACES];
};

//struttura che rappresenta un device attaccato al bus usb
struct usb_dev {
	//indirizzo del device
	u32 addr;
	//descrittore di device
	device_des dev_des;
	//descrittore di configurazione , interfaccia e endpoints
	config_des config[USB_MAXCONFIG];
	//indica se il device  configurato
	bool configured;
};

//classi usb devices
#define USB_CLASS_PER_INTERFACE		0
#define USB_CLASS_AUDIO			1
#define USB_CLASS_COMM			2
#define USB_CLASS_HID			3
#define USB_CLASS_PHYSICAL		5
#define USB_CLASS_PRINTER		7
#define USB_CLASS_MASS_STORAGE		8
#define USB_CLASS_HUB			9
#define USB_CLASS_DATA			10
#define USB_CLASS_APP_SPEC		0xfe
#define USB_CLASS_VENDOR_SPEC		0xff

//stampa sul debug di io il tipo di periferica connessa al bus
void print_type(int type)
{
	switch(type)
	{
		case USB_CLASS_AUDIO:
			dbg("AUDIO DEVICE!\n\0");
			break;
		case USB_CLASS_COMM:
			dbg("COMMUNICATION DEVICE!\n\0");
			break;
		case USB_CLASS_HID:
			dbg("HUMAN INTERFACE DEVICE!\n\0");
			break;
		case USB_CLASS_PHYSICAL:
			dbg("PHYSICAL INTERFACE DEVICE!\n\0");
			break;
		case USB_CLASS_PRINTER:
			dbg("PRINTER DEVICE!\n\0");
			break;
		case USB_CLASS_MASS_STORAGE:
			dbg("MASS STORAGE DEVICE!\n\0");
			break;
		case USB_CLASS_HUB:
			dbg("HUB DEVICE!\n\0");
			break;
		case USB_CLASS_DATA:
			dbg("DATA DEVICE!\n\0");
			break;
		case USB_CLASS_APP_SPEC:
			dbg("APPLICATION SPEC DEVICE!\n\0");
			break;
		case USB_CLASS_VENDOR_SPEC:
			dbg("VENDOR SPEC DEVICE!\n\0");
			break;
		default :
			dbg("SCONOSCIUTO!\n\0");
			break;
	}
}

//ricerca il tipo di periferica nei descrittori
void print_dev_des(usb_dev* dev)
{
	dbg("CONESSA PERIFERICA AL BUS USB DI TIPO \0");
	if(dev->dev_des.bDeviceClass!=0)
		print_type(dev->dev_des.bDeviceClass);
	else
		print_type(dev->config[0].interface[0].bInterfaceClass);
}

//Strutture dati , macro e funzioni per i TD

//Macro per indirizzare i registri interfaccia dell'uhci 

 #define W_USBCMD 0
#define   W_USBCMD_RS  0x0001 /* Run/Stop */
#define   W_USBCMD_HCRESET 0x0002 /* Host reset */
#define   W_USBCMD_GRESET  0x0004 /* Global reset */
#define   W_USBCMD_EGSM  0x0008 /* Global Suspend Mode */
#define   W_USBCMD_FGR  0x0010 /* Force Global Resume */
#define   W_USBCMD_SWDBG  0x0020 /* SW Debug mode */
#define   W_USBCMD_CF  0x0040 /* Config Flag */
#define   W_USBCMD_MAXP  0x0080 /* Max Packet (0 = 32, 1 = 64) */

 #define W_USBSTS 2
#define   W_USBSTS_USBINT  0x0001 /* USB Interrupt */
#define   W_USBSTS_ERROR  0x0002 /* USB Error Interrupt */
#define   W_USBSTS_RD  0x0004 /* Resume Detect */
#define   W_USBSTS_HSE  0x0008 /* Host System Error */
#define   W_USBSTS_HCPE  0x0010 /* Host Controller Process Error */
#define   W_USBSTS_HCH  0x0020 /* HC Halted */

 #define W_USBINTR 4
#define   W_USBINTR_TIMEOUT 0x0001 /* Timeout/CRC Interrupt Enable */
#define   W_USBINTR_RESUME 0x0002 /* Resume interrupt enable */
#define   W_USBINTR_IOC  0x0004 /* Interrupt On Complete enable */
#define   W_USBINTR_SP  0x0008 /* Short packet interrupt enable */

 #define W_FRNUM  6
 #define D_FRBASEADD 8
 #define B_SOFMOD 12

 #define W_PORTSC1 16
 #define W_PORTSC2 18
#define   W_USBPORTSC_CCS 0x0001 /* Current Connect Status ("device present") */
#define   W_USBPORTSC_CSC 0x0002 /* Connect Status Change */
#define   W_USBPORTSC_PE 0x0004 /* Port Enable/Disable */
#define   W_USBPORTSC_PEC 0x0008 /* Port Enable/Disable Change */
#define   W_USBPORTSC_LS 0x0030 /* Line Status */
#define   W_USBPORTSC_RD 0x0040 /* Resume Detect */
#define   W_USBPORTSC_LSDA 0x0100 /* Low Speed Device Attached */
#define   W_USBPORTSC_PR 0x0200 /* Port Reset */
#define   W_USBPORTSC_SUSP 0x1000 /* Suspend */

//tipo di struttura
enum type_elem {TD = 0 , QH =1};

//tipo di PID
enum PID {IN , OUT , SETUP};

//quattro doppie parole che fisicamente compongono il TD
struct _td{
	u32 link;//link FISICO all'elemento successivo
	u32 cont_st;
	u32 token;
	u32 buffer;//link fisico ai dati da trasferire
}__attribute__((aligned(32)));

//descrittore td
struct td_des{
	//parte hw del descrittore
	_td values;
	//campi dati che identificano il trasferimento
	punt link_pointer; //link LOGICO all'elemento successivo
	type_elem ty_succ;  //tipo dell'elemento successivo
	//campi dati che identifiano il transferimento
	u16 maxlen; //valore massimo possibile 1023 (usb spec.)
	u16 endpoint;//endpoint 
	u16 dev_addr;//indirizzo del device
	u8 td_pid;//packet Identification
	//dati da trasferire
	punt buffer_pointer; //puntatore LOGICO al buffer (grandezza specificata dal campo maxlen)
};

//due doppie parole che fisicamente compongono il QH
struct _qh{
	u32 link;
	u32 elem;
}__attribute__((aligned(32)));

//descrittore qh
struct qh_des{
	//parte hw del descrittore
	_qh values;
	//elemento successivo Horizontal Execution
	punt link_pointer; //successivo elemento
	type_elem ty_succ;  //tipo dell'elemento successivo
	//elemento Vertical execution
	punt elem_link_pointer; //successivo elemento
	type_elem elem_ty_succ;  //tipo dell'elemento successivo
};

//struttura che contiene il frame list vett
struct uhci_framelist {
	unsigned int frame[NUM_FRAME_POINTER];
} __attribute__((aligned(4096)));

//descrittore di dispositivo UHCI
struct uhci_des{
	//identificano il controller nel bus PCI
	int device;
	int function;
	//indirizzo base
	punt usbiobase;
	//indirizzo fisico base frame list (la frame list  una pagina di memoria 4KB 1024*4 byte)
	ind_fisico frame_list;
	//serie di variabili che indicano lo stato del controllore
	bool configured;//il controllore  configurato?(anche il root hub)
	int last_num_assigned;//ultimo numero assegnato ad una periferica usb
	int num_frames;//numero di frame che devono essere trasferiti in questa sessione
	bool active;//indica se il trasferimento dell'uhci  attivo
	//semaforo di mutua esclusione
	int mutex;
	//vettore dei frame list
	uhci_framelist* frame_list_vett;
	//struttura dello scheletro dei trasf155810662erimenti
	td_des iso;//primo elemento della catena di trasferimenti isocroni (la catena si sviluppa in linea
	//orizzontale in maniera dinamica)
	qh_des* interrupt;//primo elemento della catena di trasferimenti di tipo interrupt (in verticale)
	qh_des* control;//primo elemento della catena di trasferimenti di tipo control (in verticale)
	qh_des* bulk;//primo elemento della catena di trasferimenti di tipo bulk (in verticale)
	//puntatore alla lista di descrittori dei dispositivi collegati al bus controllato dall'UHCI
	//il driver in questione gestisce un solo device!
	usb_dev* dev_att;
};


//puntatore al descrittore uhci
static uhci_des* uhci_att=0;


//i puntatori per i descrittori dell'uhci dovrebbero essere 
//organizzati a listain modo tale che possano essere configurati pi controllori su uno
//stesso pc.Questo driver gestisce un solo controllore (il primo) tra quelli che si
//trovano sul pc. 

//funzioni di interfaccia per l'uhci

//invia il global reset all'uhci
void uhci_reset()
{
	outputw(W_USBCMD_GRESET, (ind_b)uhci_att->usbiobase + W_USBCMD);
	//devo aspettare almeno 50 millisecondi
	delay(50);
	outputw(0, (ind_b)uhci_att->usbiobase + W_USBCMD);
	//devo attendere almeno 10 millisecondi
	delay(10);
}

//invia il segnale di host controller reset
void uhci_start()
{
	short int reg=0;
	bool stato=true;
	outputw(W_USBCMD_HCRESET, (ind_b)uhci_att->usbiobase + W_USBCMD);
	do{
		inputw((ind_b)uhci_att->usbiobase + W_USBCMD,reg);
		if((reg & W_USBCMD_HCRESET)==0)
			stato=false;
	}while(stato);
}

//fa partire i trasferimenti del controllore UHCI false=STOP true=RUN
void start_uhci_transfer(bool act)
{
	short int prev;
	inputw((ind_b)uhci_att->usbiobase+W_USBCMD,prev);
	if(act==true)
		prev=prev | 1;
	else
		prev=prev & 0xfffe;
	outputw(prev,(ind_b)uhci_att->usbiobase+W_USBCMD);
	uhci_att->active=act;
}

//abilita la porta1 del root hub e invia il segnale di reset al device connesso
void port1_enable()
{
	short int prev;
	inputw((ind_b)uhci_att->usbiobase+W_PORTSC1,prev);
	prev=prev | W_USBPORTSC_PR;
	outputw(prev,(ind_b)uhci_att->usbiobase+W_PORTSC1);
	prev=prev & ~W_USBPORTSC_PR;
	delay(10);
	outputw(prev,(ind_b)uhci_att->usbiobase+W_PORTSC1);
	inputw((ind_b)uhci_att->usbiobase+W_PORTSC1,prev);
	prev=prev | W_USBPORTSC_PE;
	outputw(prev,(ind_b)uhci_att->usbiobase+W_PORTSC1);
	delay(10);
	inputw((ind_b)uhci_att->usbiobase+W_PORTSC1,prev);
	if(!(prev & W_USBPORTSC_PE))
	{
		prev=prev | W_USBPORTSC_PE;
		outputw(prev,(ind_b)uhci_att->usbiobase+W_PORTSC1);
		delay(50);
	}
}

//verifica le presenza sulla porta1 di un nuovo device
void port1_connect(bool& ris)
{
	short int prev;
	inputw((ind_b)uhci_att->usbiobase+W_PORTSC1,prev);
	prev=prev & 1;
	if(prev==0)
		ris=false;
	else
		ris=true;
}

//funzioni di debug

//stampa i campi del TD
void dbg_td(td_des* td)
{
	dbg("Campo Link:\0");
	dbg_int(td->values.link);
	dbg("Campo Status:\0");
	dbg_int(td->values.cont_st);
	dbg("Campo Token:\0");
	dbg_int(td->values.token);
	dbg("Campo Buffer:\0");
	dbg_int(td->values.buffer);
}

//stampa i campi del QH
void dbg_qh(qh_des* qh)
{
	dbg("\nCampo Link:\0");
	dbg_int(qh->values.link);
	dbg("campo Elem:\0");
	dbg_int(qh->values.elem);
}

//stampa un intero
extern "C" void dbg_int(const unsigned int num){
	char* str=(char*)mem_alloc(12);
	itostr(str,10,(long)num);
	str[10]='\n';
	str[11]='\0';
	c_term_write_n(10,str,12);
}

//stampa i campi di una catena di descrittori
void dbg_td_chain(td_des* td)
{
	int i=0;
	while(td!=0)//stampo lo stato di tutti i descrittori di trasferimento di controllo
	{
		dbg("\n");
		dbg_int(td->values.cont_st);
		//dbg_td(td);
		td=(td_des*)td->link_pointer;
		i++;
	}
}

//usata per stampare a schermo il contenuto dei registri del controllore
void uhci_dbg(){
	short int reg=0;
	dbg("\nstato:\0");
	inputw((ind_b)uhci_att->usbiobase + W_USBSTS,reg);
	dbg_int(reg);
	dbg("comando:\0");
	inputw((ind_b)uhci_att->usbiobase + W_USBCMD,reg);
	dbg_int(reg);
	dbg("port 1:\0");
	inputw((ind_b)uhci_att->usbiobase + W_PORTSC1,reg); 
	dbg_int(reg);
}


//macro di utilit per i qh e frame_list_vett

#define UHCI_PTR_TERM		0x0001
#define UHCI_PTR_QH		0x0002
#define UHCI_PTR_DEPTH		0x0004

//macro di utilit per i td

// per manipolare i bit di _td.cont_st

#define TD_CTRL_SPD  (1 << 29) /* Short Packet Detect */
#define TD_CTRL_C_ERR_MASK (3 << 27) /* Error Counter bits */
#define TD_CTRL_LS  (1 << 26) /* Low Speed Device */
#define TD_CTRL_IOS  (1 << 25) /* Isochronous Select */
#define TD_CTRL_IOC  (1 << 24) /* Interrupt on Complete */
#define TD_CTRL_ACTIVE  (1 << 23) /* TD Active */
#define TD_CTRL_STALLED  (1 << 22) /* TD Stalled */
#define TD_CTRL_DBUFERR  (1 << 21) /* Data Buffer Error */
#define TD_CTRL_BABBLE  (1 << 20) /* Babble Detected */
#define TD_CTRL_NAK  (1 << 19) /* NAK Received */
#define TD_CTRL_CRCTIMEO (1 << 18) /* CRC/Time Out Error */
#define TD_CTRL_BITSTUFF (1 << 17) /* Bit Stuff Error */
#define TD_CTRL_ACTLEN_MASK 0x7ff /* actual length, encoded as n - 1 */

#define TD_CTRL_ANY_ERROR (TD_CTRL_STALLED | TD_CTRL_DBUFERR | \
     TD_CTRL_BABBLE | TD_CTRL_CRCTIME | TD_CTRL_BITSTUFF)

#define uhci_status_bits(ctrl_sts) (ctrl_sts & 0xFE0000)
#define uhci_actual_length(ctrl_sts) ((ctrl_sts) & TD_CTRL_ACTLEN_MASK) /* 1-based */


//per manipolare il bit di validit  di _td.link

#define UHCI_TD_LINK_VALID  0x0001 /*Invalidate next element*/

//per manipolare i bit di _td.token

#define uhci_maxlen(token) ((token) << 21)
#define uhci_toggle(token) ((token) << 19)
#define uhci_endpoint(token) ((token) << 15)
#define uhci_devaddr(token) ((token) << 8)

//Valori dei PID validi

#define PID_OUT  0xe1
#define PID_IN  0x69
#define PID_SETUP 0x2d

//funzioni per aggiustare i campi dei TD

void set_td_link(td_des* des,punt succ,bool bf=false)
{
	ind_fisico ind;
	td_des* td=(td_des*)succ;
	trasforma((ind_l)(&(td->values)),ind);
	u32 add=ind;
	if(succ!=0)
	{
		des->link_pointer=succ;
		des->ty_succ=TD;
		if(bf==true)
		{
			add &= (u32)0xfffffff0;
			add |= UHCI_PTR_DEPTH;
		}
	}
	else
	{
		des->link_pointer=0;
		des->ty_succ=QH;
		add= UHCI_PTR_TERM;//va messo a 1 per invalidare il link
	
	}
	des->values.link=add;
}

void set_td_cont_st(td_des* des,bool isc_sel,bool ioc)
{
	u32 att=0;
	short int reg=0;
	inputw((ind_b)uhci_att->usbiobase+W_PORTSC1,reg);
	if((reg & W_USBPORTSC_LSDA)==1)
		att |=  (TD_CTRL_LS) | (TD_CTRL_ACTIVE)   |   (TD_CTRL_SPD)       |    (TD_CTRL_C_ERR_MASK);
	else
		att |=  (TD_CTRL_ACTIVE)   |   (TD_CTRL_SPD)       |    (TD_CTRL_C_ERR_MASK);
	if(isc_sel==false)
		att=att & ~(TD_CTRL_IOS);
	else
		att=att | (TD_CTRL_IOS);
	if(ioc==false)
		att=att & ~(TD_CTRL_IOC);
	else
		att=att | (TD_CTRL_IOC);
	des->values.cont_st=att;
}

void set_td_token(td_des* des,u32 maxlen,int data_toggle,u32 endpoint,u32 dev_addr,u32 tdpid)
{
	u32 att=0;
	att = (0 & 0x0007ff00);
	des->maxlen=maxlen-1;
	des->endpoint=endpoint;
	des->dev_addr=dev_addr;
	des->td_pid=tdpid;
	att = att | uhci_maxlen(maxlen-1);
	if(data_toggle==1)
		att=att | uhci_toggle(1);
	att = att | uhci_endpoint(endpoint);
	att = att | uhci_devaddr(dev_addr);
	att |= tdpid;
	
	des->values.token=att;
}

void set_td_buffer(td_des* des,punt buff)
{
	ind_fisico ind;
	trasforma((ind_l)(buff),ind);
	des->values.buffer=ind;
	des->buffer_pointer=buff;
}



void set_qh_link(qh_des* des,punt succ)
{
	ind_fisico ind;
	qh_des* qh=(qh_des*)succ;
	trasforma((ind_l)(&(qh->values)),ind);
	u32 add=ind;
	if(succ!=0)
	{
		des->link_pointer=succ;
		des->ty_succ=QH;
		add=add | UHCI_PTR_QH;
	}
	else
	{
		des->link_pointer=0;
		des->ty_succ=QH;
		add= UHCI_PTR_TERM;//va messo a 1 per invalidare il link
	
	}
	des->values.link=add;
}

void set_qh_elem(qh_des* des,punt succ,type_elem next)
{
	if(next==QH)
	{
		ind_fisico ind;
		qh_des* qh=(qh_des*)succ;
		trasforma((ind_l)(&(qh->values)),ind);
		u32 add=ind;
		if(succ!=0)
		{
			des->elem_link_pointer=succ;
			des->elem_ty_succ=next;
			if(next==QH)
				add=add | UHCI_PTR_QH;
		}
		else
		{
			des->elem_link_pointer=0;
			des->elem_ty_succ=QH;
			add= UHCI_PTR_TERM;//va messo a 1 per invalidare il link
		
		}
		des->values.elem=add;
	}
	else
	{
		ind_fisico ind;
		td_des* td=(td_des*)succ;
		trasforma((ind_l)(&(td->values)),ind);
		u32 add=ind;
		if(succ!=0)
		{
			des->elem_link_pointer=succ;
			des->elem_ty_succ=next;
			if(next==QH)
				add=add | UHCI_PTR_QH;
		}
		else
		{
			des->elem_link_pointer=0;
			des->elem_ty_succ=QH;
			add= UHCI_PTR_TERM;//va messo a 1 per invalidare il link
	
		}
		des->values.elem=add;
	}
}

//ritorna lo stato di un TD
int get_td_status(td_des* td)
{
	u32 status=0;
	status = (td->values.cont_st >> 16) & 0xff;
}

//libera un TD o una catena di TD
void td_free(td_des* td)
{
	while(td!=0)
	{
		td_des* del=td;
		td=(td_des*)td->link_pointer;
		mem_free(td->buffer_pointer);
		mem_free(del);
	}
}


//USB control transfer strutture dati e funzioni

#define USBREQ_SET_ADDRESS		0x0500
#define USBREQ_SET_CONFIGURATION	0x0900
#define USBREQ_GET_DESCRIPTOR		0x0680


#define DES_DEVICE_VALUE	0x0100
#define DES_CONFIGURATION_VALUE	0x0200
#define DES_INTERFACE_VALUE	0x0402
#define DES_ENDPOINT_VALUE	0x0500

//struttura per le richieste standard
//8 byte lunghezza standard
struct control_data{ 
	u16 req;
	u16 value;
	u16 index;
	u16 length;
}__attribute__((aligned(8)));


//crea una catena di tdrestituendone i puntatori al primo e all'ultimo
td_des* create_td_chain(u32 max_packet_length,void* buffer,u32 length,u32 tpid,u32 dev,u32 endpoint,td_des*& last,int& toggle)
{
	int num=(length/max_packet_length);
	if((length % max_packet_length)!=0)
		num++;
	td_des* ini;
	td_des* fin;
	ind_b punt=(ind_b)buffer;
	int i=0;
	for(;i<num;i++)
	{
		length=length-max_packet_length;
		td_des* des=alloc_td();
		memset(des,0,sizeof(td_des));
		set_td_cont_st(des,false,false);
		set_td_token(des,max_packet_length,(i+toggle)%2,endpoint,dev,tpid);
		set_td_buffer(des,punt);
		set_td_link(des,0);
		if(i==0)
			ini=des;
		else
		{
			if(i!=(num-1))
				set_td_link(fin,des,true);
			else
				set_td_link(fin,des);
		}
		fin=des;
		punt+=max_packet_length;
	}
	toggle=num%2; 
	last=fin;
	return ini;
}

//crea una lista di td restituendo un puntatore al primo
td_des* create_control_message(u16 action, u16 dev, u16 value,td_des*& stat,punt buffer,u32 size)
{
	int toggle;
	td_des* last;
	control_data* cont=(control_data*)mem_alloc(sizeof(control_data));
	memset(cont,0,sizeof(control_data));
	cont->req=action;
	td_des* td_base=/*(td_des*)mem_alloc(sizeof(td_des))*/alloc_td();
	memset(td_base,0,sizeof(td_des));
	switch(action)
	{
		case USBREQ_SET_ADDRESS:
			cont->value=value;
			cont->index=0;
			cont->length=0;
			set_td_link(td_base,0);
			set_td_cont_st(td_base,false,false);
			set_td_token(td_base,8,0,0,0,PID_SETUP);
			set_td_buffer(td_base,(ind_b)cont);
			last=td_base;
			break;
		case USBREQ_SET_CONFIGURATION:
			cont->value=value;
			cont->index=0;
			cont->length=0;
			set_td_link(td_base,0);
			set_td_cont_st(td_base,false,false);
			set_td_token(td_base,8,0,0,dev,PID_SETUP);
			set_td_buffer(td_base,(ind_b)cont);
			last=td_base;
			break;
		case USBREQ_GET_DESCRIPTOR:
			cont->value=value;
			cont->index=0;
			cont->length=size;
			set_td_link(td_base,0);
			set_td_cont_st(td_base,false,false);
			set_td_token(td_base,8,0,0,dev,PID_SETUP);
			set_td_buffer(td_base,(ind_b)cont);
			toggle=1;
			td_des* int_ch=create_td_chain(8,buffer,size,PID_IN,dev,0,last,toggle);
			set_td_link(td_base,int_ch,true);
			break;
	}
	if(action==USBREQ_GET_DESCRIPTOR)
	{
		stat=alloc_td();
		memset(stat,0,sizeof(td_des));
		set_td_link(stat,0);
		set_td_cont_st(stat,false,true);
		set_td_token(stat,0,toggle,0,dev,PID_OUT);
		set_td_buffer(stat,0);
		set_td_link(last,stat,true);
	}
	else
	{
		stat=alloc_td();
		memset(stat,0,sizeof(td_des));
		set_td_link(stat,0);
		set_td_cont_st(stat,false,true);
		set_td_token(stat,0,1,0,dev,PID_IN);
		set_td_buffer(stat,0);
		set_td_link(last,stat);
	}
	return td_base;
}

//assegna un indirizzo al device
void set_address(usb_dev* dev)
{
	td_des* stat;
	td_des* td=create_control_message(USBREQ_SET_ADDRESS,0,dev->addr,stat,0,0);
	u32 st_ini=stat->values.cont_st;
	set_qh_elem(uhci_att->control,td,TD);
	int i=0;
	while(((TD_CTRL_ACTIVE & stat->values.cont_st)!=0)||(get_td_status(stat)!=0)){
		i++;
		if(i>TIMEOUT)//timeout esaurito
		{
			#ifdef DEBUG_USB
			uhci_dbg();
			#endif 
			dbg("TIMEOUT SCATTATO!\n\0");
			break;
		}
	}//controllo di programma
	if(get_td_status(stat)!=0)
	{
		dbg("NON RIESCO AD ASSEGNARE AL DEVICE UN INDIRIZZO!\n\0");
		#ifdef DEBUG_USB
		dbg_td_chain(td);
		#endif
	}
	else
		dbg("INDIRIZZO ASSEGNATO AL DEVICE...\n\0");
	set_qh_elem(uhci_att->control,0,TD);
	/*mem_free(td);
	mem_free(stat);*/
	td_free(td);
}

//assegna una configurazione al device
void set_configuration(usb_dev* dev,u32 conf)
{
	td_des* stat;
	td_des* td=create_control_message(USBREQ_SET_CONFIGURATION,dev->addr,conf,stat,0,0);
	u32 st_ini=stat->values.cont_st;
	set_qh_elem(uhci_att->control,td,TD);
	int i=0;
	while(((TD_CTRL_ACTIVE & stat->values.cont_st)!=0)||(get_td_status(stat)!=0)){
		i++;
		if(i>TIMEOUT)//timeout esaurito
		{
			#ifdef DEBUG_USB
			uhci_dbg();
			#endif 
			dbg("TIMEOUT SCATTATO!\n\0");
			break;
		}
	}//controllo di programma
	if(get_td_status(stat)!=0)
	{
		dbg("NON RIESCO AD ASSEGNARE LA CONFIGURAZIONE AL DEVICE!\n\0");
		#ifdef DEBUG_USB
		dbg_td_chain(td);
		#endif
	}
	else
		dbg("CONFIGURAZIONE ASSEGNATA AL DEVICE...\n\0");
	set_qh_elem(uhci_att->control,0,TD);
	td_free(td);
}

// //riempie i descrittori del device
void get_device_desc(usb_dev* dev)
{
	td_des* stat;
	td_des* 
	//recupero il descrittore del device
	td=create_control_message(USBREQ_GET_DESCRIPTOR,dev->addr,DES_DEVICE_VALUE,stat,&(uhci_att->dev_att->dev_des),sizeof(device_des));
	u32 st_ini=stat->values.cont_st;
	//dbg_td_chain(td);
	set_qh_elem(uhci_att->control,td,TD);
	int i=0;
	int retr=5;
	while(((TD_CTRL_ACTIVE & stat->values.cont_st)!=0)||(get_td_status(stat)!=0)){
		i++;
		if(i>TIMEOUT)//timeout esaurito
		{
			#ifdef DEBUG_USB
			uhci_dbg();
			#endif 
			dbg("TIMEOUT SCATTATO!\n\0");
			break;
		}
	}//controllo di programma
	if(get_td_status(stat)!=0)
	{
		dbg("NON RIESCO A OTTENERE IL DESCRITTORE DEL DEVICE!\n\0");
		#ifdef DEBUG_USB
		dbg_td_chain(td);
		#endif
	}
	else
		dbg("DESCRITTORE DEVICE RECUPERATO...\n\0");
	set_qh_elem(uhci_att->control,0,TD);
	td_free(td);
	delay(50);
	//recupero il descrittore della configurazione (al massimo ce n' una sola) , di interfaccia e di endpoint
	td=create_control_message(USBREQ_GET_DESCRIPTOR,dev->addr,DES_CONFIGURATION_VALUE,stat,&(uhci_att->dev_att->config),sizeof(config_des));
	st_ini=stat->values.cont_st;
	set_qh_elem(uhci_att->control,td,TD);
	i=0;
	while(((TD_CTRL_ACTIVE & stat->values.cont_st)!=0)||(get_td_status(stat)!=0)){
		i++;
		if(i>TIMEOUT)//timeout esaurito
		{
			#ifdef DEBUG_USB
			uhci_dbg();
			#endif 
			dbg("TIMEOUT SCATTATO!\n\0");
			break;
		}
	}//controllo di programma
	if(get_td_status(stat)!=0)
	{
		dbg("NON RIESCO A OTTENERE I DESCRITTORI GENERICI!\n\0");
		#ifdef DEBUG_USB
		dbg_td_chain(td);
		#endif
	}
	else
		dbg("DESCRITTORE CONFIGURAZIONE , INTERFACCIA E ENDPOINT RECUPERATI...\n\0");
	set_qh_elem(uhci_att->control,0,TD);
	td_free(td);	
}

//effettua la configurazione di un device appena attaccato
//la waitforconnectusb deve chiamarla subito dopo aver individuato
//il cambio di stato del bus dovuto al plug di un nuovo device
void configure_device(){
	dbg("CONFIGURO DEVICE....\n\0");
	//abilito la porta
	port1_enable();
	//alloco il descrittore
	uhci_att->dev_att=(usb_dev*)mem_alloc(sizeof(device_des));
	//assegno il primo indirizzo disponibile al device
	uhci_att->last_num_assigned++;
	uhci_att->dev_att->addr=uhci_att->last_num_assigned;
	//assegno il primo indirizzo disponibile
	set_address(uhci_att->dev_att);
	//attendo 10 ms per far si che il device acquisisca l'indirizzo
	delay(10);
	//assegno la prima configurazione al device
	set_configuration(uhci_att->dev_att,1);
	//attendo 10 ms per far si che il device acquisisca la configurazione
	delay(10);
	
	uhci_att->dev_att->configured=true;
	//recupero i descrittori	
	get_device_desc(uhci_att->dev_att);
	//stampo il tipo di device attaccato
	print_dev_des(uhci_att->dev_att);
}

//funzioni per il funzionamento dei devices attaccati

//funzioni per l'uso di una stampante

//14 byte per preambolo alla stampa
#define	PREAMBOLO_STAMPA	0x0D0A0A0A0A0A0A20202020202020
//1 byte per la chiusura della spampa
#define CHIUSURA_STAMPA		0x0C

//se in device attaccato  una stampante stampa il testo nel buffer
extern "C" void print_text(char* buf,int lun)
{
	dbg("INIT PRINTING...\n\0");
	char* buffer=(char*)mem_alloc(lun + 15);
	buffer[0]=0x0D;
	buffer[1]=0x0A;
	buffer[2]=0x0A;
	buffer[3]=0x0A;
	buffer[4]=0x0A;
	buffer[5]=0x0A;
	buffer[6]=0x0A;
	buffer[7]=0x20;
	buffer[8]=0x20;
	buffer[9]=0x20;
	buffer[10]=0x20;
	buffer[11]=0x20;
	buffer[12]=0x20;
	buffer[13]=0x20;
	for(int i=14;i<(lun+14);i++)
		buffer[i]=buf[i-14];
	buffer[lun+14]=0x0C;
	int toggle=0;
	td_des* last;
	td_des* first=create_td_chain(64,buffer,lun+15,PID_OUT,(uhci_att->dev_att->addr),2,last,toggle);
	u32 st_ini=last->values.cont_st;
	set_qh_elem(uhci_att->bulk,first,TD);
	while((TD_CTRL_ACTIVE & last->values.cont_st)!=0){}//controllo di programma
	set_qh_elem(uhci_att->bulk,0,TD);
	td_free(first);
}

//ricerca nello spazio pci il primo controllore usb e ne restituisce l'indirizzo base nello spazio di io

int pci_usb_find(int& dev,int& fun)
{
	int bus=0;
	dev=0;
	fun=0;
	if(pci_find(USB_CLASS_CODE,USB_SUB_CLASS_CODE,USB_PROG_IF,bus,dev,fun))
	{
		dbg("TROVATO CONTROLLORE USB NELLO SPAZIO PCI!\n\0");
		int iospace=0;
		for(int i=4;i<10;i++)
		{
			iospace=pci_interr(bus,dev,fun,i);
			if((iospace & 1)==1)//ho un indirizzo valido?
			{
				
				//dbg_int((iospace & 0xfffffffc));
				return (iospace & 0xfffffffc);
			}
		}
	}
	dbg("CONTROLLORE USB NON PRESENTE!");
	return 0;
}

//questa funzione ricerca il controller UHCI all'interno del bus pci e ritorna il suo indirizzo
void usb_find(ind_b& usbbas,int& dev,int& fun)
{
	dbg("RICERCO NELLO SPAZIO PCI UN CONTROLLORE USB...\n\0");
	usbbas=(ind_b)pci_usb_find(dev,fun);
}



//inizializza lo scheletro dei descrittori di trasferimento con le tre QH per i tre tipi di trasferimento
//le tre QH devono formare un loop per ottimizzare il tempo!
void uhci_skel_init()
{
	dbg("CONFIGURO SKEL....\n\0");
	uhci_att->interrupt=alloc_qh();
	uhci_att->control=alloc_qh();
	uhci_att->bulk=alloc_qh();
	set_qh_link(uhci_att->interrupt,(punt)(uhci_att->control));
	set_qh_link(uhci_att->control,(punt)(uhci_att->bulk));
	set_qh_link(uhci_att->bulk,(punt)(uhci_att->interrupt));//bandwidth reclamation LOOP
	//set_qh_link(&uhci_att->bulk,0);
	set_qh_elem(uhci_att->interrupt,0,TD);
	set_qh_elem(uhci_att->control,0,TD);
	set_qh_elem(uhci_att->bulk,0,TD);
	ind_fisico ind;
	trasforma((ind_l)(&uhci_att->interrupt->values),ind);
	for(int i=0;i<1024;i++)
		uhci_att->frame_list_vett->frame[i]=ind | 2;
}

//inizializza registri e strutture dati dell'uhci
void usb_init()
{
	dbg("SCHERMATA DI LOG DELLO SPAZIO DI I/O:\n\0");
	ind_b usbbas;
	int dev,fun;
	usb_find(usbbas,dev,fun);
	if(usbbas!=0)
	{
		
		dbg("CONFIGURO UHCI....\n\0");
		//alloco lo spazio
		uhci_att=(uhci_des*)mem_alloc(sizeof(uhci_des));
		memset(uhci_att,0,sizeof(uhci_des));
		//inizializzo i campi del descrittore uhci
		uhci_att->usbiobase=usbbas;
		uhci_att->device=dev;
		uhci_att->function=fun;
		//creo il framelist
		uhci_att->frame_list_vett=alloc_uhci_framelist();
		//cerco l'indirizzo fisico del frame list
		trasforma((ind_l)uhci_att->frame_list_vett,uhci_att->frame_list);
		//la struttura  inizializzata ad 1 per invalidare tutte le voci
		memset(uhci_att->frame_list_vett->frame,1,(sizeof(u32)*NUM_FRAME_POINTER));
		//inizializzo i registri dell'uhci
		uhci_reset();
		uhci_start();
		//il numero zero  gi assegnato di default (spec. USB)
		uhci_att->last_num_assigned=0;
		//inizializzo lo scheletro delle strutture dati dell'uhci
		uhci_skel_init();
		//abilito le interruzioni sul controllore uhci
		outputw(0x000f,(ind_b)uhci_att->usbiobase+W_USBINTR);
		//inizializzo il frame base register
		outputl(uhci_att->frame_list,(ind_b)uhci_att->usbiobase + D_FRBASEADD);
		//segnalo che il controllore  configurato
		outputw(0, (ind_b)uhci_att->usbiobase + W_FRNUM);
		outputw(/*W_USBCMD_SWDBG | */W_USBCMD_RS | W_USBCMD_CF | W_USBCMD_MAXP,(ind_b)uhci_att->usbiobase + W_USBCMD);
		uhci_att->configured=true;
	}
	else
		//non c' alcun controller UHCI non creo il descrittore il descrittore nullo  il segnale per 
		//le altri parti del driver che non  presente alcun controller , altre azioni vanno quindi annullate
		return ;
	
}






//questa funzione si preoccupa di attendere la connessione di un dispositivo ad una delle porte del root
//hub. Questa funzione deve essere mandata in esecuzione dall'usbd il demone che controlla l'usb.
//essa  bloccante e si risveglia automaticamente dopo n secondi per effetto della chiamata sleep!
//restituisce 0 se non  stato connesso nulla e 1 o 2 se vi  stata connessione sulla porta 1 o 2
extern "C" void wait_for_connect_usb(int& port,int& type)//da esportare sul lato user!parte fondamentale dell'usbd
{
	if(uhci_att==0)
	{
		usb_init();
		if(uhci_att==0)
		{
			port=-1;
			return;
		}
	}
	#ifdef DEBUG_USB
	uhci_dbg();
	#endif
	bool ris=false;
	port1_connect(ris);
	if(ris==true)
	{
		if(uhci_att->dev_att==0)
			configure_device();//configuro il device assegnandogli un address e una configurazione
		type=uhci_att->dev_att->dev_des.bDeviceClass;
		if(type==0)
			type=type=uhci_att->dev_att->config[0].interface[0].bInterfaceClass;
		port=1;
		return;
	}
	port=0;
	//la porta numero due dl root hub resta per ora morta.
	return;
	
}





////////////////////////////////////////////////////////////////////////////////
//                 INIZIALIZZAZIONE DEL SOTTOSISTEMA DI I/O                   //
////////////////////////////////////////////////////////////////////////////////

// inizializza i gate usati per le chiamate di IO
//
extern "C" void fill_io_gates(void);

// eseguita in fase di inizializzazione ad interruzioni disabilitate
//
extern "C" void io_init(void)
{	
	fill_io_gates();
	term_init();
	com_init();
	//hd_init();
	//dma_init();
	//fdc_init();
}







